#Typing imports
import typing as typ

#External imports
import os

#Internal Imports
from dbgenPython.main import load_jobs

user = os.environ['USER']

#####################
# Sync Clusters
# -------------------

def sync_suncat()->None:
    print('Syncing /nfs/slac/g/suncatfs/ksb/share/jobs to /scratch/users/ksb/share/suncat_jobs_copy/ ...')
    os.system('rsync -rtqvu --perms --chmod=777 --omit-dir-times %s@suncatls1.slac.stanford.edu:/nfs/slac/g/suncatfs/ksb/share/jobs /scratch/users/ksb/share/suncat_jobs_copy/ --delete'%user)
def sync_nersc()->None:
    print('Syncing /global/cscratch1/sd/krisb/share/jobs to /scratch/users/ksb/share/nersc_jobs_copy/ ...')
    os.system('rsync -rtqvu --perms --chmod=777 --omit-dir-times krisb@cori.nersc.gov:/global/cscratch1/sd/krisb/share/jobs /scratch/users/ksb/share/nersc_jobs_copy/ --delete')

#####################
# Load Database
# -------------------
def load_jobs_catalog(parallel : bool = True)-> None:
    load_jobs(catalog  = True
             ,parallel = parallel
             ,only     = """
                         populate_catalog_job

                         update_job_metadata
                         populate_relax_job
                         populate_subjobs

                         populate_relax_calc
                         populate_calc_other
                         kptden

                         populate_struct_traj_atom_cell
                         update_system_type
                         populate_substructs

                         pop_struct_catalog
                         pop_ads_catalog
                         pop_bulk_struct_catalog
                         pop_surf_struct_catalog
                         """)



# Typing Modules
import typing as typ
if typ.TYPE_CHECKING:
    import pandas as pd  # type: ignore

# External modules
# import sqlite3
import subprocess,math,os,sys,json,warnings, sqlite3
import sql                                        # type: ignore
from sql.aggregate import Sum                     # type: ignore
import ase                                        # type: ignore
import prettytable                                # type: ignore
from ase.visualize import view                    # type: ignore
from ase.data import (atomic_masses, atomic_names # type: ignore
                     ,chemical_symbols, covalent_radii
                     ,reference_states)
from ase.build import bulk, molecule  # type: ignore

# Internal modules
from catalog.misc.utilities import sub_binds
from catalog.misc.print_parse import ask, read_storage_json
from catalog.misc.atoms import make_atoms, json_to_traj
from catalog.misc.sql_utils import *
import catalog.misc.sql_utils as sql_utils
"""
Primary functions for interfacing with main database, $SCRATCH/db/data.db
"""
#################################################################################
# Constants
###########

##########################################################
# QUERY CLASS
##########################################################


realDB = ConnectInfo(host='mysql01.stanford.edu'  # 'g-suncat-suncatdata.sudb.stanford.edu'
                     , port=3306, user='gsuncatsuncatd', passwd='BCe8HzyXCA-ekD!!', db='g_suncat_suncatdata')


class Query(object):
    """
    Read from a SQLite database
    """

    def __init__(self
                ,cols       : typ.List[sql.Column]     = [tbl_relax_job.job_id]
                ,constraints: typ.List[typ.Any]        = []
                ,table      : sql.Table                = tbl_relaxfinal
                ,order      : typ.Optional[sql.Column] = None
                ,group      : typ.Optional[sql.Column] = None
                ,limit      : typ.Optional[int]        = None
                ,connectinfo: ConnectInfo              = realDB
                ,verbose    : bool                     = False
                ,distinct   : bool                     = False
                 ) -> None:

        self.constraints = constraints
        self.cols = cols
        self.table = table
        self.order = order
        self.group = group
        self.limit = limit
        self.connectinfo = connectinfo
        self.verbose = verbose

    def query(self
             ,table      : typ.Optional[sql.Table]   = None
             ,constraints: typ.Optional[list]        = None
             ,cols       : typ.Optional[list]        = None
             ,connectinfo: typ.Optional[ConnectInfo] = None
              ) -> list:
        """ [STRING] -> CONSTRAINT -> [[sqloutput]] """
        if cols is None:
            cols = self.cols
        if table is None:
            table = self.table
        if constraints is None:
            constraints = self.constraints
        if connectinfo is None:
            connectinfo = self.connectinfo

        command = self._command(constraints=constraints,
                                cols=cols, table=table)
        if self.verbose == True:
            sub_binds(command)
        return sqlselect(conn=connectinfo, q=tuple(command)[0].replace('"', '`'), binds=tuple(command)[1])

    def query_dict(self, table: typ.Optional[sql.Table] = None, constraints: typ.Optional[list] = None, cols: typ.Optional[list] = None, connectinfo: typ.Optional[ConnectInfo] = None
                   ) -> list:
        """ [STRING] -> CONSTRAINT -> [{colname:colval}] """
        if cols is None:
            cols = self.cols
        if table is None:
            table = self.table
        if constraints is None:
            constraints = self.constraints
        if connectinfo is None:
            connectinfo = self.connectinfo

        command = self._command(constraints=constraints,
                                cols=cols, table=table)
        if self.verbose == True:
            sub_binds(command)

        return select_dict(conn=connectinfo, q=tuple(command)[0].replace('"', '`'), binds=tuple(command)[1])

    def _command(self
                ,constraints: list
                ,cols       : list
                ,table      : sql.Table
                 ) -> typ.Any:
        if len(cols) == 1 and '*' == str(cols[0]):
            cols = []
        self.last_command = table.select(
            *cols, where=And(constraints), order_by=self.order, limit=self.limit, group_by=self.group)
        return self.last_command

    def query_col(self, col: sql.Column, constraints: typ.Optional[list] = None, table: typ.Optional[sql.Table] = None
                  ) -> list:
        assert isinstance(
            col, sql.Column), 'Please supply a single column (You probably used a singleton list)'
        query_output = self.query(
            cols=[col], constraints=constraints, table=table)
        return [row[0] for row in query_output]

    def make_atoms(self) -> typ.List[ase.Atoms]:
        """returns a list of atoms corresponding to each row in the database"""
        assert 'struct' in table_names(
            self.table), 'The query is not joined to the structure table'
        raw_jsons = self.query_col(tbl_struct.raw)
        return [json_to_traj(json_curr) for json_curr in raw_jsons]

    def view(self
             ) -> None:
        """
        View with ase-gui the initial or final atoms object for a given job - specified by a constraint
        """
        stordirs = self.query_col(tbl_job.stordir)
        output = zip(self.make_atoms(), stordirs)  # type: ignore
        for i, (atoms_obj, storage_directory) in enumerate(output):
            print('#%d/%d stored at: %s' %
                  (i + 1, len(stordirs), storage_directory))
            view(atoms_obj)
            if (i + 1 == len(stordirs)) or 'q' in input():
                break

    def query1(self, col: sql.Column) -> typ.Any:
        output = self.query_col(col)
        if len(output) != 1:
            raise ValueError(
                'query1 did not return one result: ' + str(output))
        return output[0]

    def any_query(self, anycol: sql.Column
                  ) -> bool:
        """
        A column must be specified that is in the table
        """
        return len(self.query_col(anycol)) > 0

    def col_names(self) -> typ.List[str]:
        self.table_names = table_names(self.table)
        col_names = []  # type: list
        for table_name in self.table_names:
            command = "\
                      SELECT `COLUMN_NAME`\
                      FROM `INFORMATION_SCHEMA`.`COLUMNS`\
                      WHERE `TABLE_SCHEMA`='g_suncat_suncatdata'\
                          AND `TABLE_NAME`='%s';\
                      " % (table_name)
            col_names += list(zip(*
                                  sqlselect(conn=self.connectinfo, q=command)))[0]
            # output = [table_name+'.'+col_name for col_name in col_names]
        return list(set(col_names))

    # def dataframe(self)->"pd.DataFrame":
    #     """
    #
    #     """
    #     import pandas as pd
    #     col_names       = map(str,self.cols)
    #     list_of_columns = zip(*self.query())                    # convert rows to columns
    #     df = pd.DataFrame(dict(zip(col_names,list_of_columns))) # match with column names
    #     return df
    #
    # def csv(self,fileout : str ='out.csv') -> None:
    #     import csv
    #     col_names = map(str,self.cols)
    #     rows      = self.query()
    #     with open(fileout, 'w+') as csvfile:
    #         writer = csv.writer(csvfile, delimiter='$',
    #                                 quotechar='|', quoting=csv.QUOTE_MINIMAL)
    #         writer.writerow(col_names)
    #         for row in rows: writer.writerow(row)

    # def query_parent_cols(self
    #                      ,cols              : list
    #                      ,added_constraints : list = []
    #                      ,warn_if_no_parent : bool = True
    #                      ) -> list:
    #     """
    #     Allows the user to query columns from the
    #     If the final_table is being used a
    #     """
    #     assert self.table in [final_table], 'querying parent columns has only been implemented for the final_table in sql'
    #     if None in self.query_col(job.parent) and warn_if_no_parent:
    #         warnings.simplefilter('always', UserWarning)
    #         warnings.warn('Some rows in this query have no parents', UserWarning, stacklevel = 1)
    #
    #     new_constraints = self.constraints + added_constraints
    #     if not isinstance(cols, list):
    #         return self.query_col(cols, constraints = new_constraints, table = parent_child_table)
    #     return self.query(cols = cols, constraints = new_constraints, table = parent_child_table)

    # def ptable(self)->None:
    #
    #     command         = self._command(self.constraints,self.cols,self.table)
    #     connection      = sqlite3.connect(self.db_path)
    #     cursor          = connection.cursor()
    #     if self.verbose: print(tuple(command))
    #     cursor.execute(*tuple(command))
    #     mytable         = prettytable.from_db_cursor(cursor) # type: ignore
    #     cursor.close();connection.close()# type: ignore
    #     print(mytable)
    #     print('Results: %d'%len(mytable._rows))


class AdsEnergyQuery(object):
    """
    Specific query for querying Adsorption Energies

    inner_constraints are applied before the group concat is applied. This can
    (and should) be used to choose the ref_scheme_id, calc comparison parameters,
    and other constraints to limit which ads_triples are legitamate

    outer_constraints are applied to the last select statement, you have access
    to the default columns:
    (bare_id, complex_id,name, bare_name (a job_name), complex_name (a job_name)
    e_ads, job_dict)
    As well as any other columns you add to the inner inner_columns_dict

    inner_columns_dict is in the format {column_alias: table.column}
    The possible tables to query from are listed below along with their variable
    name (see catalog.misc.sql_utils for source code for each variable):

    Description             Table           Varname
    N/A                     ads_triple      T
    Bare finaltraj          finaltraj       BF
    Complex finaltraj       finaltraj       CF
    Bare relax_job          relax_job       BR
    Complex relax_job       relax_job       CR
    Bare job                job             BJ
    Complex job             job             CJ
    Bare struct             struct          BS
    Complex struct          struct          CS
    Bare calc_other         calc_other      BC
    Complex calc_other      calc_other      CC
    ads_eng                 ads_eng         A
    adsorbate job           job             AJ
    Adsorbate calc_other    calc_other      AC
    """

    def __init__(self, inner_constraints: typ.List[typ.Any] = [], outer_constraints: typ.List[typ.Any] = [], inner_columns_dict: typ.Dict[sql.Column, str] = {}, order: typ.Optional[sql.Column] = None, group: typ.Optional[sql.Column] = None, limit: typ.Optional[int] = None, connectinfo: ConnectInfo = realDB, verbose: bool = False
                 ) -> None:

        self.inner_constraints = inner_constraints
        self.outer_constraints = outer_constraints
        self.inner_columns_dict = inner_columns_dict
        self.order = order
        self.group = group
        self.limit = limit
        self.connectinfo = connectinfo
        self.verbose = verbose

    def query(self) -> typ.List[tuple]:
        command = self._command()
        if self.verbose == True:
            sub_binds(command)
        return sqlselect(conn=self.connectinfo, q=tuple(command)[0].replace('"', '`'), binds=tuple(command)[1])

    def query_dict(self) -> typ.List[dict]:
        command = self._command()
        if self.verbose == True:
            sub_binds(command)
        return select_dict(conn=self.connectinfo, q=tuple(command)[0].replace('"', '`'), binds=tuple(command)[1])

    def _command(self) -> sql.Select:
        inner_columns = [As(val, key)
                         for key, val in self.inner_columns_dict.items()]
        inner_select = sql_utils.tbl_eads.select(
            *sql_utils.eads_columns_inner + inner_columns, where=self.inner_constraints, group_by=sql_utils.eads_group_by_inner)
        outer_select_cols = [inner_select.bare_id, inner_select.complex_id, inner_select.name, inner_select.bare_name, inner_select.complex_name, As(inner_select.delta_e_surf - Sum(
            inner_select.energy_norm * inner_select.coef * inner_select.number), 'e_ads'), As(CON('[', GROUPCONCAT(CON('(', inner_select.element_id, ',[', inner_select.jobs, '])')), ']'), 'job_dict')]

        eads_group_by_outer = [inner_select.bare_id, inner_select.complex_id]
        added_outer_select_cols = [sql.Column(
            inner_select, name) for name in self.inner_columns_dict.keys()]

        outer_select = inner_select.select(
            *outer_select_cols + added_outer_select_cols, group_by=eads_group_by_outer)
        self.last_command = outer_select.select()
        return self.last_command

    def make_atoms(self
                   ) -> typ.List[dict]:
        tmp_dict = self.inner_columns_dict
        self.inner_columns_dict = {}
        self.inner_columns_dict['ads_raw'] = CON(
            '[', GROUPCONCAT(CON(AS.raw)), ']')
        self.inner_columns_dict['bare_raw'] = BS.raw
        self.inner_columns_dict['complex_raw'] = CS.raw
        query_output = self.query_dict()
        output_list = []
        for row in query_output:
            row_dict = {}
            for raw_name in self.inner_columns_dict.keys():
                if raw_name == 'ads_raw':
                    row[raw_name] = json.loads(row[raw_name])
                    row_dict[raw_name] = [json_to_traj(
                        json.dumps(raw_curr)) for raw_curr in row[raw_name]]
                else:
                    row_dict[raw_name] = json_to_traj(str(row[raw_name]))
            output_list.append(row_dict)
        self.inner_columns_dict = tmp_dict
        return output_list

    def view(self, view_refs: bool = False
             ) -> None:
        output_atom_dicts = self.make_atoms()
        output_row_dicts = self.query_dict()
        for i, (atoms_dict, row_dict) in enumerate(zip(output_atom_dicts, output_row_dicts)):
            print('#%d/%d' % (i + 1, len(output_atom_dicts)))
            print(self._row_dict_summary(row_dict))
            flattened_list = [atoms_dict['bare_raw'],
                              atoms_dict['complex_raw']]
            if view_refs:
                view(atoms_dict['ads_raw'] + flattened_list)
            else:
                view(flattened_list)
            if (i + 1 == len(output_atom_dicts)) or 'q' in input():
                break

    @staticmethod
    def _row_dict_summary(row_dict: dict, summary_keys: list = ['bare_name', 'complex_name', 'e_ads', 'bare_id', 'complex_id']
                          ) -> str:

        lines = ['{0} : {1}'.format(key, val) if key in summary_keys else '' for key,
                 val in row_dict.items() if key in summary_keys]
        return '\n'.join(lines)


###################
# Interface with DB
# ------------------
def sqlite_execute(db_path: str, sqlcommand: str, binds: list = []
                   ) -> None:
    """
    Execute a SQL statement that modifies the database
    """
    assert sqlcommand.lower()[:6] != 'select'
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()
        cursor.execute('PRAGMA foreign_keys = ON')
        cursor.execute(sqlcommand, binds)
        output_array = cursor.fetchall()
        conn.commit()


def sqlite_select(db_path: str, sqlcommand: str, binds: list = [], mk_dict: bool = False
                  ) -> list:
    """
    Select and return a list of tuples
    """
    def dict_factory(cursor, row):  # type: ignore
        return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}

    assert sqlcommand.lower()[:6] == 'select'

    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()

        if mk_dict:
            cursor.row_factory = dict_factory

        cursor.execute(sqlcommand, binds)
        output_array = cursor.fetchall()
    return output_array


def select_dict(conn: ConnectInfo, q: str, binds: list = []
                ) -> typ.List[dict]:
    #print('selecting dict ',q)
    with conn.mk_conn(mk_dict=True) as cxn:  # type: ignore
        if 'group_concat' in q.lower():
            cxn.execute("SET SESSION group_concat_max_len = 100000")
        cxn.execute(q, args=binds)
        return cxn.fetchall()


def sqlselect(conn: ConnectInfo, q: str, binds: list = []
              ) -> typ.List[tuple]:
    # print('\n\nSQLSELECT ',q)#,binds)
    with conn.mk_conn() as cxn:  # type: ignore
        if 'group_concat' in q.lower():
            cxn.execute("SET SESSION group_concat_max_len = 100000")
        cxn.execute(q, args=binds)
        return cxn.fetchall()


def sqlexecute(conn: ConnectInfo, q: str, binds: list = []
               ) -> None:
    print('\n\nSQLEXECUTE ', q)  # ,binds)
    with conn.mk_conn() as cxn:  # type: ignore
        cxn.execute("SET SESSION auto_increment_offset = 1")
        cxn.execute("SET SESSION auto_increment_increment = 1")
        cxn.execute(q, args=binds)


def sqlexecutemany(conn: ConnectInfo, q: str, binds: typ.List[list]
                   ) -> None:
    print('executemany Q: ', q, binds)
    with conn.mk_conn() as cxn:  # type: ignore
        cxn.executemany(q, args=binds)


###################
# MODIFY Functions
###################
def iif(a, b, c): return b if a else c  # type: ignore


###########
# def sqlexecute(sqlcommand, binds = [],db_path=dbreal,row_factory=None,hacks = []): # type: ignore
#     assert sqlcommand.lower()[:6] in ['create','select','insert','delete','alter ','update','drop t'], "Sql Query weird "+sqlcommand
#     connection = sqlite3.connect(db_path,timeout=30)
#     if 'sqrt' in sqlcommand: connection.create_function('SQRT',1,math.sqrt)
#
#     for h in hacks: sqlcommand = h(sqlcommand) # apply hacks to string
#
#     cursor     = connection.cursor()
#     cursor.row_factory = row_factory
#     cursor.execute(sqlcommand, binds)
#     output_array = cursor.fetchall()
#
#
#     if sqlcommand.lower()[:6] != 'select': connection.commit()
#     connection.close()
#     return output_array
#
# def sqlexecutemany(sqlcommand, binds = [],db_path=dbreal): # type: ignore
#     assert sqlcommand.lower()[:6] in ['create','select','insert','delete','alter ','update'], "Sql Query weird "+sqlcommand
#     connection  = sqlite3.connect(db_path,timeout=60)
#     cursor      = connection.cursor()
#     cursor.executemany(sqlcommand, binds)
#     if 'select' not in sqlcommand.lower(): connection.commit()
#     cursor.close() # type: ignore
#     connection.close()
#     return

# def updateDB(set_column : str
#             ,id_column  : str
#             ,ID         : int
#             ,new_value  : typ.Any
#             ,table      : str
#             ,db_path    : str     =dbreal
#             ) -> None:
#     """
#     """
#     sqlexecute("UPDATE {0} SET {1}= ? WHERE {2} = ?".format(table,set_column,id_column),[new_value,ID],db_path=db_path)# type: ignore
#
# def drop_table(tableName : str
#               ,db_path   : str =dbreal
#               ) -> None:
#     """
#     """
#     print('dropping table ',tableName)
#     try: sqlexecute("DROP TABLE %s"%tableName,db_path=db_path) # type: ignore
#     except Exception as e: print('\tdrop_table exception: ',e)
# def deleteRow(tableName : str
#              ,idColumn  : str
#              ,ID        : int
#              ,db_path   : str = '/scratch/users/ksb/share/suncatdata.db'
#              ) -> None:
#     """
#     """
#     sqlexecute("DELETE FROM {0} WHERE {1} = {2}".format(tableName,idColumn,ID),db_path=db_path) # type: ignore
#
# def addCol(colname   : str
#           ,coltype   : str
#           ,tableName : str
#           ,db_path   : str
#           ) -> None:
#     sqlexecute("ALTER TABLE {0} ADD COLUMN {1} {2}".format(tableName,colname,coltype),db_path=db_path)# type: ignore

###################
# Creating/deleting
###################
# def get_max(tab: str, col: str, db_path: str
#             ) -> int:
#     """
#     """
#     cmd = 'SELECT {0} FROM {1} ORDER BY {0} DESC LIMIT 1'
#     args = [col, tab]
#     return sqlexecute(cmd.format(*args), db_path=db_path)[0][0]  # type: ignore
#
#
# def removeSuncatDataDB(db_path: str ='/scratch/users/ksb/share/temp.db') -> None:
#     """
#     """
#     print("removing ", db_path)
#     if os.path.exists(db_path):
#         os.remove(db_path)
#
#
# def createSuncatDataDB(db_path: str ='/scratch/users/ksb/share/temp.db') -> int:
#     print('creating DB at ', db_path)
#
#     fill_element(db_path)
#
#     sqlexecute(('CREATE TABLE atom        (id     integer primary key'  # type: ignore
#                 + ',atoms_id    integer not null'
#                 + ',ind         integer not null'
#                 + ',number      integer not null'
#                 + ',x           numeric not null'
#                 + ',y           numeric not null'
#                 + ',z           numeric not null'
#                 + ',constrained integer not null'
#                 + ',magmom      numeric not null'
#                 #+',tag         integer'
#                 # don't want 2 rows describing same index of same Atoms object
#                 + ',UNIQUE (atoms_id,ind)'
#                 + ',FOREIGN KEY(number) REFERENCES element(id)'
#                 + ',FOREIGN KEY(atoms_id) REFERENCES atoms(id))'), db_path=db_path)
#
#     sqlexecute(("CREATE TABLE cell     (id     integer primary key"  # type: ignore
#                 + ',ax numeric not null'
#                 + ',ay numeric not null'
#                 + ',az numeric not null'
#                 + ',bx numeric not null'
#                 + ',by numeric not null'
#                 + ',bz numeric not null'
#                 + ',cx numeric not null'
#                 + ',cy numeric not null'
#                 + ',cz numeric not null'
#                 + ',UNIQUE(ax,ay,az,bx,by,bz,cx,cy,cz))'), db_path=db_path)
#
#     sqlexecute(("CREATE TABLE atoms     (id     integer primary key"  # type: ignore
#                 + ',cell_id integer not null'
#                 + ',FOREIGN KEY(cell_id) REFERENCES cell(id))'), db_path=db_path)
#
#     sqlexecute(("CREATE TABLE calc        (id  integer primary key"  # type: ignore
#                 + ',dftcode  varchar not null'
#                 + ',xc       varchar'
#                 + ',pw       integer'
#                 + ',kpts     varchar'
#                 + ',fmax     numeric'
#                 + ',psp      varchar'
#                 + ',econv    numeric'
#                 + ',xtol     numeric'
#                 + ',strain   numeric'
#                 + ',dw       integer'
#                 + ',sigma    numeric'
#                 + ',nbands   integer'
#                 + ',mixing   numeric'
#                 + ',nmix     integer'
#                 + ',gga      varchar'
#                 + ',luse_vdw integer'
#                 + ',zab_vdw  numeric'
#                 + ',nelmdl   integer'
#                 + ',gamma    integer'
#                 + ',dipol    varchar'
#                 + ',algo     varchar'
#                 + ',ibrion   integer'
#                 + ',prec     varchar'
#                 + ',ionic_steps integer'
#                 + ',lreal    varchar'
#                 + ',lvhar    integer'
#                 + ',diag     varchar'
#                 + ',spinpol  integer'
#                 + ',dipole   integer'
#                 + ',maxstep  integer'
#                 + ',delta    numeric'
#                 + ',mixingtype varchar'
#                 + ',bonded_inds varchar'
#                 + ',energy_cut_off numeric'
#                 + ',step_size numeric'
#                 + ',springs varchar'
#                 + ',UNIQUE(dftcode,xc,pw,kpts,fmax,psp,econv,xtol,strain,dw,sigma,nbands,mixing,nmix,gga,luse_vdw,zab_vdw,nelmdl,gamma,dipol,algo,ibrion,prec,ionic_steps'
#                 + ',lreal,lvhar,diag,spinpol,dipole,maxstep,delta,mixingtype,bonded_inds,energy_cut_off,step_size,springs))'), db_path=db_path)
#
#     sqlexecute("CREATE TABLE job          (id                integer primary key"  # type: ignore
#                + ',deleted           integer not null'
#                + ',user              varchar not null'
#                + ',timestamp         numeric not null'
#                + ',working_directory varchar not null unique'
#                + ',storage_directory varchar not null unique'
#                + ',job_name          varchar not null '
#                + ',guess_jobtype     varchar not null '
#                + ',calc_id           integer not null '
#                + ',initatoms         integer not null '
#                + ',finalatoms        integer not null '
#                + ',FOREIGN KEY(calc_id)    REFERENCES calc  (id)'
#                + ',FOREIGN KEY(initatoms)  REFERENCES atoms (id)'
#                + ',FOREIGN KEY(finalatoms) REFERENCES atoms (id))', db_path=db_path)
#
#     # sqlexecute("CREATE TABLE keyword      (id integer primary key"  # type: ignore
#     #                                    +',job_id integer not null unique'
#     #                                    +',FOREIGN KEY(job_id) REFERENCES job (id))',db_path=db_path)
#
#     sqlexecute("CREATE TABLE composition (id integer primary key"  # type: ignore
#                + ',element_id integer not null '
#                + ',atoms_id   integer not null '
#                + ',has        integer not null '
#                + ',UNIQUE (element_id,atoms_id)'
#                + ',FOREIGN KEY(element_id) REFERENCES element (id)'
#                + ',FOREIGN KEY(atoms_id) REFERENCES atoms (id))', db_path=db_path)
#
#     fill_element(db_path)
#     add_refstoich(db_path)
#     add_refeng(db_path)
#     os.system('chmod 777 ' + db_path)
#     return 1


###################
# Supplementary
# ----------------
# def add_refeng(db_path: str =dbreal) -> None:
#     sqlexecute("CREATE TABLE refeng      (id integer primary key"  # type: ignore
#                + ',element_id integer not null '
#                + ',calc_id integer not null '
#                # this isn't property of calculator so it must be appended here
#                + ',kptden_x numeric not null '
#                # this isn't property of calculator so it must be appended here
#                + ',kptden_y numeric not null '
#                + ',UNIQUE (element_id,calc_id)'
#                + ',FOREIGN KEY(element_id) REFERENCES element (id)'
#                + ',FOREIGN KEY(calc_id) REFERENCES calc (id))', db_path=db_path)
#
#
# def add_refstoich(db_path: str =dbreal) -> None:
#     sqlexecute("CREATE TABLE refstoich (id integer primary key"  # type: ignore
#                + ',reference_element integer'
#                + ',component_element integer'
#                + ',component_weight  numeric'
#                + ',total_components  integer'
#                + ',FOREIGN KEY(reference_element) REFERENCES element (id)'
#                + ',FOREIGN KEY(component_element) REFERENCES element (id))', db_path=db_path)
#     cols = 'reference_element,component_element,component_weight,total_components'
#     for i in range(1, 119):
#         if i == 6:
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [i, i, 1, 3], db_path)  # type: ignore
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [6, 8, -1, 3], db_path)  # type: ignore
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [6, 1, 2, 3], db_path)  # type: ignore
#         elif i == 8:
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [i, i, 1, 2], db_path)  # type: ignore
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [8, 1, -2, 2], db_path)  # type: ignore
#         else:
#             sqlexecute('INSERT into refstoich (%s) VALUES (?,?,?,?)' %
#                        cols, [i, i, 1, 1], db_path)  # type: ignore
#
#
# def add_element(db_path: str =dbreal) -> None:
#     sqlexecute(("CREATE TABLE element   (id     integer primary key"   # type: ignore
#                 + ',symbol  varchar not null'
#                 + ',name    varchar not null'
#                 + ',mass    numeric not null'
#                 + ',radius  numeric not null'
#                 + ',reference_phase varchar'
#                 + ',reference_spacegroup integer '
#                 + ',reference_pointgroup varchar'
#                 # brian's columns
#                 + ',atomic_number                integer'
#                 + ',atomic_weight                numeric'
#                 + ',abundance_crust              numeric'
#                 + ',abundance_sea                numeric'
#                 + ',atomic_radius                numeric'
#                 + ',atomic_volume                numeric'
#                 + ',atomic_weight_uncertainty    numeric'
#                 + ',boiling_point                numeric'
#                 + ',covalent_radius_bragg        numeric'
#                 + ',covalent_radius_cordero      numeric'
#                 + ',covalent_radius_pyykko       numeric'
#                 + ',covalent_radius_pyykko_double   numeric'
#                 + ',covalent_radius_pyykko_triple   numeric'
#                 + ',covalent_radius_slater   numeric'
#                 + ',density                  numeric'
#                 + ',dipole_polarizability    numeric'
#                 + ',econf                    numeric'
#                 + ',electron_affinity    numeric'
#                 + ',en_allen             numeric'
#                 + ',en_ghosh             numeric'
#                 + ',en_pauling           numeric'
#                 + ',evaporation_heat     numeric'
#                 + ',fusion_heat          numeric'
#                 + ',gas_basicity         numeric'
#                 + ',geochemical_class    varchar'
#                 + ',goldschmidt_class    varchar'
#                 + ',group_id             integer'
#                 + ',heat_of_formation    numeric'
#                 + ',is_radioactive       bool'
#                 + ',lattice_constant     numeric'
#                 + ',lattice_structure    varchar'
#                 + ',melting_point        numeric'
#                 + ',metallic_radius      numeric'
#                 + ',metallic_radius_c12  numeric'
#                 + ',period               integer'
#                 + ',proton_affinity      numeric'
#                 + ',thermal_conductivity numeric'
#                 + ',vdw_radius   numeric)'), db_path=db_path)
#
#
# def fill_element(db_path: str = dbreal) -> None:
#
#     add_element(db_path)
#
#     # Constants
#     # ---------
#     sym_dict = {'fcc': 225, 'diamond': 227, 'bcc': 229, 'hcp': 194, 'sc': 221, 'cubic': None, 'rhombohedral': None,
#                 'tetragonal': None, 'monoclinic': None, 'orthorhombic': None, 'diatom': 'D*h', 'atom': 'K*h'}
#
#     ase_cols = ['symbol', 'name', 'mass', 'radius', 'reference_phase',
#                 'reference_spacegroup', 'reference_pointgroup']
#     mendeleev_cols = ['atomic_number', 'atomic_weight', 'abundance_crust', 'abundance_sea', 'atomic_radius', 'atomic_volume',
#                       'atomic_weight_uncertainty', 'boiling_point', 'covalent_radius_bragg', 'covalent_radius_cordero', 'covalent_radius_pyykko',
#                       'covalent_radius_pyykko_double', 'covalent_radius_pyykko_triple', 'covalent_radius_slater', 'density', 'dipole_polarizability',
#                       'econf', 'electron_affinity', 'en_allen', 'en_ghosh', 'en_pauling', 'evaporation_heat', 'fusion_heat', 'gas_basicity', 'geochemical_class',
#                       'goldschmidt_class', 'group_id', 'heat_of_formation', 'is_radioactive', 'lattice_constant', 'lattice_structure', 'melting_point',
#                       'metallic_radius', 'metallic_radius_c12', 'name', 'period', 'proton_affinity', 'symbol', 'thermal_conductivity', 'vdw_radius']
#     cols = ase_cols + mendeleev_cols
#
#     with open(os.environ['HOME'] + '/CataLog/data/element.json', 'r') as f:
#         eledicts = json.load(f)
#
#     binds = []
#     qmarks = ','.join(['?'] * len(cols))
#
#     # Main Loop
#     # ----------
#     for i in range(1, 119):
#         s, n, m, r = [x[i] for x in [chemical_symbols,
#                                      atomic_names, atomic_masses, covalent_radii]]
#         if reference_states[i] is None:
#             rp, rsg, rpg = None, None, None
#         elif 'a' in reference_states[i]:
#             rp, rpg = 'bulk', None
#             rsg = sym_dict[reference_states[i]['symmetry']]
#             if i == 50:
#                 rsg = 227  # ase doesn't think tin is diamond
#         else:
#             rp, rsg = 'molecule', None
#             rpg = sym_dict[reference_states[i]['symmetry']]
#         ase_binds = [s, n, m, r, rp, rsg, rpg]
#
#         m_binds = [eledicts[i - 1].get(m_c) for m_c in mendeleev_cols]
#
#         binds.append(ase_binds + m_binds)
#
#     sqlexecutemany('INSERT INTO element (%s) VALUES (%s)' %
#                    (','.join(cols), qmarks), binds, db_path)  # type: ignore
#
#     return None
#
# ####################
# # MODIFYING STORAGE
# ####################
#
#
# def modify_storage_directory(constraint, file_name, key, value, verbose=True, check=True, dict_loc='params'):  # type: ignore
#
#     def writeOnSherOrSlac(pth, file_text):  # type: ignore
#         if 'scratch' in pth:
#             open(pth, 'w').close()
#             with open(pth, 'w') as f:
#                 return f.write(file_text)
#         elif 'nfs' in pth:
#             writer = subprocess.Popen(['ssh', '{0}@suncatls1.slac.stanford.edu'.format(
#                 user), 'cat -> %s' % (pth)], stdin=subprocess.PIPE, stderr=subprocess.PIPE)
#             writer.stdin.write(file_text)
#             return 1
#         else:
#             raise NotImplementedError('New cluster? path =' + pth)
#
#     # Define protected keys that cannot be changed
#     protected_params = ['kpts', 'jobkind', 'xc', 'maxstep', 'nbands', 'dftcode', 'dwrat', 'kptden', 'finaltraj_pckl', 'finalmagmom_pckl', 'spinpol',
#                         'forces_pckl', 'finalcell_pckl', 'fmax', 'mixing', 'dw', 'psp', 'econv', 'pw', 'nmix', 'finalpos_pckl', 'raw_energy', 'inittraj_pckl', 'sigma']
#     assert key not in protected_params, 'Cannot change job results or calculation parameters in finished jobs'
#     # Define acceptable files to change
#     assert file_name in [
#         'result.json', 'params.json'], 'Can only modify param.json and result.json'
#     # Grab directories that match constraints
#     additionalConstraint = USER_(user)  # type: ignore
#     query = Query(verbose=True, constraints=AND(
#         additionalConstraint, *constraint))
#     directories = query.query_col(STORDIR)
#
#     def modify_dict(curr_dict, key, value, dict_loc):  # type: ignore
#         if not key in curr_dict.keys():
#             if check and ask('Do you want to add new key "{0}" to '.format(key) + file_name):
#                 print('Modifying file: {0}'.format(directory + file_name))
#                 print('Changing {0} parameter: {1}; From {2} to {3}'.format(
#                     dict_loc, key, curr_dict.get(key, "empty"), value))
#                 curr_dict[key] = value
#                 writeOnSherOrSlac(directory + file_name,
#                                   json.dumps(curr_params))   # type: ignore
#         else:
#             print('Modifying file: {0}'.format(directory + file_name))
#             print('Changing {0} parameter: {1}; From {2} to {3}'.format(
#                 dict_loc, key, curr_dict.get(key, "empty"), value))
#             curr_dict[key] = value
#         return curr_dict
#
#     # Iterate through directories and change the value matching the key provided
#     for directory in directories:
#         curr_params = read_storage_json(
#             directory + file_name)  # Get Current Params
#         # Need to check if key in keyword table as they don't exist in curr_params.keys()
#         if dict_loc in ['kwarg', 'project']:
#             curr_dict = json.loads(curr_params.get(
#                 dict_loc)) if dict_loc in curr_params.keys() else {}  # type: ignore
#             assert isinstance(
#                 curr_dict, dict), 'Current parameter value for dict_loc is not a dictionary'  # type: ignore
#             if not curr_dict.get(key) == value:
#                 new_dict = modify_dict(
#                     curr_dict, key, value, dict_loc)  # type: ignore
#                 curr_params[dict_loc] = json.dumps(
#                     new_dict)             # type: ignore
#                 writeOnSherOrSlac(directory + file_name,
#                                   json.dumps(curr_params))  # type: ignore
#         elif dict_loc == 'params':
#             # type: ignore
#             if not curr_params.get(key) == value:
#                 curr_params = modify_dict(
#                     curr_params, key, value, dict_loc)      # type: ignore
#                 writeOnSherOrSlac(directory + file_name,
#                                   json.dumps(curr_params))   # type: ignore
#         else:
#             raise NotImplementedError('Specified dict_loc: {0} has not been made before, please add to the list {1}'.format(
#                 dict_loc, ['kwarg', 'project', 'params']))



#Typing imports
import typing as typ

#External imports
import subprocess,os

#Internal Imports




class SSHClient(object):
    def __init__(self
                ,hostname : str
                ,username : str
                )->None:
        """
        Object for interacting across ssh terminals

        Parameters
        ----------
        hostname : str
            hostname of the server to be interacted with. (i.e. login.stanford.edu)
        username : str
            username of the current user
        """

        self.hostname = hostname
        self.username = username

        def scp(self):
            return

        def exec_command(self):
            return

        def _check_connection(self):
            return



#External imports
import sys

major_version, minor_version = sys.version_info[:2]
if major_version < 3 or minor_version < 6:
    raise Exception("Python 3.6 or a more recent version is required.")

import venv # type: ignore
import os, shutil

def build_venv(env_dir):
    if not os.path.exists(env_dir):
        os.mkdir(env_dir)
    venv.create(os.path.join(env_dir,'.env'),with_pip = True,symlinks = True)
    shell = os.getenv('SHELL','')
    if 'bash' in shell or 1:
        rcpath = os.path.join(env_dir,'.env/bin/activate')
    elif 'tcsh' in shell:
        rcpath = os.path.join(env_dir,'.env/bin/activate.csh')
    else:
        print("Shell must be tcsh or bash, you are using {}".format(shell))
    set_user_enviromental_vars(rcpath)
    python_exec = os.path.join(env_dir,'.env/bin/python')
    os.system("{} -m pip install pip --upgrade --no-cache-dir".format(python_exec))
    os.system("{} -m pip install numpy --upgrade --no-cache-dir".format(python_exec))
    os.system("{} -m pip install CC-CataLog --upgrade --no-cache-dir".format(python_exec))
    add_fireworks_folders()
    if 'bash' in shell:
        print("""
        !!!!!!IMPORTANT!!!!!
        NEED TO DO ONE MORE THING
        PLEASE RUN THIS COMMAND
        echo export CATALOG_LOC={}>>~/.bashrc
        """.format(os.environ['CATALOG_LOC']))
    elif 'tcsh' in shell:
        print("""
        !!!!!!IMPORTANT!!!!!
        NEED TO DO ONE MORE THING
        PLEASE RUN THIS COMMAND
        echo setenv CATALOG_LOC {}>>~/.cshrc
        """.format(os.environ['CATALOG_LOC']))

def set_user_enviromental_vars(rcpath):
    user_keys = {'USER'               : "Generic username used as default for new clusters"
                ,'SHERLOCK2_USERNAME' : "Username for Sherlock2 cluster"
                ,'SUNCAT_USERNAME'    : "Username for Suncat Cluster"
                ,'HOSTNAME'               : "Name of current cluster (sherlock, suncat, local)"
                ,'LAUNCHPAD_YAML'     : "path to launchpad_yaml for fireworks"
                }
    set_user_keys = {} # type: dict
    if rcpath.endswith('.csh') and 0:
        with open(rcpath,'a') as f:
            f.write("#####################\n")
            f.write("#Catalog enviromental variables\n")
            f.write("#####################\n")
            f.write("setenv PYTHONPATH {}:$PYTHONPATH\n".format(env_dir))
            for var, desc in user_keys.items():
                print('#####################')
                print(var+': '+desc)
                user_value = input("Please Enter your value for env-var, {}: ".format(var))
                os.environ[var] = user_value
                print("value for {} set to {}".format(var,user_value))
                f.write("setenv {} {}\n".format(var,user_value))
            print('#####################')
    else:
        with open(rcpath,'a') as f:
            f.write("#####################\n")
            f.write("#Catalog enviromental variables\n")
            f.write("#####################\n")
            f.write("export PYTHONPATH={}:$PYTHONPATH\n".format(env_dir))
            for var, desc in user_keys.items():
                print('#####################')
                print(var+': '+desc)
                user_value = input("Please Enter your value for env-var, {}: ".format(var))
                os.environ[var] = user_value
                print("value for {} set to {}".format(var,user_value))
                f.write("export {}={}\n".format(var,user_value))
            print('#####################')

def add_fireworks_folders():
    #Make all the necessary folders
    safe_mkdir(os.path.join(env_dir,'fireworks'))
    safe_mkdir(os.path.join(env_dir,'fireworks','jobs'))
    safe_mkdir(os.path.join(env_dir,'fireworks','FW_CONFIG_DIR'))
    safe_mkdir(os.path.join(env_dir,'fireworks','logs'))

    #Make a FW_config.yaml
    with open(os.path.join(env_dir,'fireworks','FW_config.yaml'),'w') as f:
        f.write('ALWAYS_CREATE_NEW_BLOCK: True')

    #Fill the FW_CONFIG_DIR with my_launchpad, my_fworker, and my_qadapter
    #Copy the launchpad_yaml
    shutil.copy(os.environ['LAUNCHPAD_YAML'],os.path.join(env_dir,'fireworks','FW_CONFIG_DIR','my_launchpad.yaml'))

    #Write the my_fworker.yaml
    with open(os.path.join(env_dir,'fireworks','FW_CONFIG_DIR','my_fworker.yaml'),'w') as f:
        f.write("name: {}\n".format(get_cluster()))
        f.write("category: ''\n")
        f.write("query: '{}'\n")

    #Write the my_qadapter.yaml
    with open(os.path.join(env_dir,'fireworks','FW_CONFIG_DIR','my_qadapter.yaml'),'w') as f:
        f.write("_fw_name: CommonAdapter\n")
        f.write("_fw_q_type: LoadSharingFacility\n")
        f.write("queue: suncat\n")

    #Copy the launcher function to the fireworks folder
    with open(os.path.join(env_dir,'fireworks','launcher.sh'),'w') as f:
        f.write("""
        #!/bin/bash
        source $CATALOG_LOC/.env/bin/activate
        $CATALOG_LOC/.env/bin/qlaunch -c $CATALOG_LOC/fireworks/FW_CONFIG_DIR --logdir $CATALOG_LOC/fireworks/logs -r rapidfire
        """)
    os.chmod(os.path.join(env_dir,'fireworks','launcher.sh'),0755)

def safe_mkdir(dir_path):
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)

def get_cluster():# type: ignore
    """Get the name of the cluster this command is being executed on"""
    hostname = os.environ['HOSTNAME'] # type: ignore
    if      'sh'     in hostname: return 'sherlock2'
    elif    'gpu-15' in hostname: return 'sherlock2'
    elif    'su'     in hostname: return 'suncat' #important to distinguish suncat2 and 3?
    elif    'kris'   in hostname: return 'kris'
    elif    'local'  in hostname: return 'local'
    else: raise ValueError("was not able to identify the cluster from hostname = %s"%hostname)


if __name__ == '__main__':
    env_dir = input('Please enter the path you would like to put the CataLog enviroment\n(Please choose an empty or non-existent folder): ')
    env_dir = os.path.abspath(env_dir)
    os.environ['CATALOG_LOC'] = env_dir
    build_venv(env_dir)



#Typing imports
import typing as typ

#External imports
import qdarkstyle                                 # type: ignore
import sqlparse                                   # type: ignore
import time, logging, subprocess, os
from ase.visualize import view                    # type: ignore
from PyQt5 import QtCore, QtGui, QtWidgets, QtSql # type: ignore

#Internal Imports
import catalog.fw.incomplete as inc
from catalog.misc.sql_utils import *
import catalog.jobs.cluster as cluster
from catalog.misc.utilities import merge_dicts, get_cluster, read_into_temp, get_hostname_from_fworker
from catalog.catalog_config import USER,SHERLOCK2_USERNAME,SUNCAT_USERNAME


class QPlainTextEditLogger(logging.Handler):
    def __init__(self, parent : typ.Any) -> None:
        super().__init__()
        self.widget = QtWidgets.QPlainTextEdit(parent)
        self.widget.setReadOnly(True)

    def emit(self, record : typ.Any)-> None:
        msg = self.format(record)
        self.widget.appendPlainText(msg)

class LaunchForm(QtWidgets.QDialog):
    def __init__(self
                ,fwid_list : typ.List[int]
                ,parent    : typ.Any      = None
                )->None:
        super(LaunchForm, self).__init__()
        self.parent = parent
        self.setupUi()

        self.fwid_list    = fwid_list
        self.old_dicts    = {fwid:inc.lpad.get_fw_dict_by_id(fwid)['spec'] for fwid in fwid_list}
        self.modify_dicts = {fwid:{} for fwid in fwid_list} # type: dict
        self.load_list()
        self.load_table()
        self._connect_signals()

    def _connect_signals(self) -> None:
        #Connect push buttons
        self.push_save.clicked.connect(self._save)
        self.push_undo.clicked.connect(self._undo)
        self.push_view.clicked.connect(self._view)
        self.push_archive.clicked.connect(self._archive)
        self.push_fizzle.clicked.connect(self._fizzle)
        self.push_defuse.clicked.connect(self._defuse)
        self.push_launch.clicked.connect(self.launch)

        self.list.itemSelectionChanged.connect(self.load_table)

    def setupUi(self) -> None:
        """
        Create the UI for the QueryGUI object
        """
        header_font = QtGui.QFont()
        header_font.setPointSize(15)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.finish)
        self.createFormGroupBox()
        #Layour table and list
        self.table = QtWidgets.QTableWidget()
        self.table.setColumnCount(2)
        self.table.setHorizontalHeaderItem(0, QtWidgets.QTableWidgetItem())
        item = self.table.horizontalHeaderItem(0)
        item.setText('Key')
        self.table.setHorizontalHeaderItem(1, QtWidgets.QTableWidgetItem())
        item = self.table.horizontalHeaderItem(1)
        item.setText('Value')
        font = QtGui.QFont()
        font.setPointSize(20)

        self.list = QtWidgets.QListWidget()
        self.list.setSizePolicy(QtWidgets.QSizePolicy.Minimum,QtWidgets.QSizePolicy.Minimum)

        #Make saving button
        self.push_save = QtWidgets.QPushButton('Save')
        self.push_save.setFont(font)
        #Make Undo button
        self.push_undo = QtWidgets.QPushButton('Undo All Changes')
        self.push_undo.setFont(font)
        #Make Undo button
        self.push_view = QtWidgets.QPushButton('View')
        self.push_view.setFont(font)
        #Make Archive button
        self.push_archive = QtWidgets.QPushButton('Archive All')
        self.push_archive.setFont(font)
        #Make Fizzle button
        self.push_fizzle = QtWidgets.QPushButton('Fizzle All')
        self.push_fizzle.setFont(font)
        #Make Defuse button
        self.push_defuse = QtWidgets.QPushButton('Defuse All')
        self.push_defuse.setFont(font)
        #Make Defuse button
        self.push_launch = QtWidgets.QPushButton('Launch All')
        self.push_launch.setFont(font)

        #Setup logger
        self.logTextBox = QPlainTextEditLogger(self)
        # # You can format what is printed to text box
        self.logTextBox.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logging.getLogger().handlers = []
        logging.getLogger().addHandler(self.logTextBox)
        # You can control the logging level
        logging.getLogger().setLevel(logging.DEBUG)

        #Lay stuff out
        table_widget = QtWidgets.QGroupBox()
        table_layout = QtWidgets.QHBoxLayout()
        table_layout.addWidget(self.list)
        table_layout.addWidget(self.table)
        table_layout.addWidget(self.logTextBox.widget)

        table_widget.setLayout(table_layout)

        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.addWidget(self.formGroupBox)
        mainLayout.addWidget(table_widget)
        mainLayout.addWidget(self.push_save)
        mainLayout.addWidget(self.push_undo)
        mainLayout.addWidget(self.push_view)
        self.frame = QtWidgets.QFrame()
        fizzle_archive_layout = QtWidgets.QHBoxLayout()
        fizzle_archive_layout.addWidget(self.push_fizzle)
        fizzle_archive_layout.addWidget(self.push_archive)
        fizzle_archive_layout.addWidget(self.push_defuse)
        fizzle_archive_layout.addWidget(self.push_launch)
        self.frame.setLayout(fizzle_archive_layout)
        mainLayout.addWidget(self.frame)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)
        logging.info('Welcome!')

    def createFormGroupBox(self)->None:
        self.formGroupBox = QtWidgets.QGroupBox("Launch Form")
        layout = QtWidgets.QFormLayout()

        #Build Walltime spinner
        self.walltime_spin = QtWidgets.QSpinBox()
        self.walltime_spin.setMinimum(0); self.walltime_spin.setMaximum(144);
        self.walltime_spin.setValue(20);
        layout.addRow(QtWidgets.QLabel("Walltime:"), self.walltime_spin)

        #build cluster
        self.cluster_combo = QtWidgets.QComboBox()
        [self.cluster_combo.insertItem(i,name) for i,name in enumerate(list(cluster.cluster_dict.keys())+['Same Cluster']+['Fizzle'])]
        self.cluster_combo.setCurrentIndex(len(cluster.cluster_dict.keys()))
        layout.addRow(QtWidgets.QLabel("Cluster:"), self.cluster_combo)

        self.nodes_spin = QtWidgets.QSpinBox()
        self.nodes_spin.setMinimum(1); self.nodes_spin.setMaximum(6);
        self.nodes_spin.setValue(1);
        layout.addRow(QtWidgets.QLabel("Nodes:"), self.nodes_spin)
        self.formGroupBox.setLayout(layout)

    def load_list(self)->None:
        self.list.clear()
        for fwid in self.fwid_list:
            self.list.addItem(str(fwid))
        self.list.setCurrentRow(0)

    def load_table(self)->None:
        banned_keys = ['inittraj']
        self.table.clearContents()
        if len(self.fwid_list)>0:
            self.table.setRowCount(0)
            banned_keys = ['inittraj']
            selected_fwid = int(self.list.currentItem().text())
            fwid_dict = inc.lpad.get_fw_dict_by_id(selected_fwid)['spec']
            queueadapter = fwid_dict['_queueadapter']
            params = fwid_dict['params']
            merged_dict = merge_dicts([params])
            for key,val in merged_dict.items():
                if not key in banned_keys:
                    currentRowCount = self.table.rowCount()
                    self.table.insertRow(currentRowCount)
                    self.table.setItem(currentRowCount,0, QtWidgets.QTableWidgetItem(str(key)))
                    if self.modify_dicts[selected_fwid].get(key,None):
                        val = self.modify_dicts[selected_fwid].get(key,None)
                    self.table.setItem(currentRowCount,1, QtWidgets.QTableWidgetItem(str(val)))
            self.table.resizeColumnsToContents()

    def _save(self)->None:
        selected_fwid = int(self.list.currentItem().text())
        fw_spec       = inc.lpad.get_fw_dict_by_id(selected_fwid)['spec']
        old_dict      = merge_dicts([fw_spec['params']])
        for i_row in range(self.table.rowCount()):
            key = self.table.item(i_row,0).text()
            val = self.table.item(i_row,1).text()
            old_type = type(old_dict[key])
            if not old_type == str and not old_type == type(None):
                if old_type == list or old_type == bool:
                    val = eval(val)
                else:
                    val = old_type(val)
            if old_type == type(None):
                pass
            elif not val == old_dict[key]:
                logging.info('Changed {} for fwid {} from {} to {}'.format(key,selected_fwid,old_dict[key],val))
                self.modify_dicts[selected_fwid][key] = val

    def _undo(self)->None:
        self.modify_dicts = {fwid:{} for fwid in self.fwid_list}
        self.load_table()

    def _defuse(self)->None:
        for fwid in self.fwid_list:
            inc.lpad.defuse_fw(fwid)

    def _fizzle(self)->None:
        for fwid in self.fwid_list:
            state = inc.lpad.get_fw_dict_by_id(fwid)['state']
            if state in ['RESERVED','RUNNING']:
                inc.fizzle(fwid,force_cancel=True)
            else:
                logging.error('fwid {} can\'t be fizzled it has state: {}'.format(fwid,state))

    def _archive(self)->None:
        for fwid in self.fwid_list:
            inc.archive_by_fwid(fwid)

    def _view(self)->None:
        """
        View a traj file in the launch directory of the selected fireworkself.
        Looks in the current launch directory of the selected fwid and finds the
        non empty traj files. Prompts the user to specify which one they would
        like to view.
        """
        import tempfile
        #Get the selected fwid
        selected_fwid = int(self.list.currentItem().text())
        #Find the launch_dir and cluster of that job
        launch_dir    = self.old_dicts[selected_fwid]['_launch_dir']
        fworker       = self.old_dicts[selected_fwid]['_fworker']
        hostname      = get_hostname_from_fworker(fworker)
        #Find which traj files are not empty in that launch directory
        command       = 'ssh {}@{} find "{}/*.traj" -not -empty -print'.format(USER,hostname,launch_dir)
        output        = subprocess.check_output(command, shell=True)
        traj_files    = output.decode().split('\n')[:-1]
        base_names    = [os.path.basename(file) for file in traj_files]
        #Open up a dialog box to ask the user which file to view
        dialogbox     = QtWidgets.QInputDialog()
        item, ok      = dialogbox.getItem(self, 'Viewer','Select a traj to view:',base_names,0,False)
        #View that traj file
        if item and ok:
            local_path       = read_into_temp(traj_files[base_names.index(item)])
            output           = subprocess.check_output('python -c "from catalog.misc.atoms import smart_view; smart_view(\'{}\')"'.format(local_path),shell=True)
            #Remove the temporary file
            os.remove(local_path)


    def launch(self) -> None:
        clusters_to_launch = []
        walltime    = self.walltime_spin.value()
        clust_str   = self.cluster_combo.currentText()

        logging.info('Rerunning fwids = {}'.format(self.fwid_list))
        nodes       = int(self.nodes_spin.value())
        for fwid in self.fwid_list:
            state   = inc.lpad.get_fw_dict_by_id(fwid)['state']
            if state   == 'DEFUSED' :
                inc.lpad.reignite_fw(fwid)
            elif state == 'RUNNING' :
                logging.error('Can\'t rerun fwid {}, it has state {}'.format(fwid,state))
                continue

            fw_spec = inc.lpad.get_fw_dict_by_id(fwid)['spec']
            fworker   = fw_spec['_fworker']

            if clust_str == 'Same Cluster':
                to_cluster = cluster.cluster_dict[fworker]
            else:
                to_cluster  = cluster.cluster_dict[clust_str]

            if not fworker == to_cluster.name:
                inc.switchCluster(fwid, from_cluster_str = fworker, to_cluster_str = clust_str)

            inc.modify(fwid,'_queueadapter', lambda x: to_cluster.qfunc(walltime,nodes))
            inc.modify_p(fwid,self.modify_dicts[fwid])
            if len(self.modify_dicts[fwid])>0:
                logging.info('Param keys modified for fwid {}: {}'.format(fwid,list(self.modify_dicts[fwid].keys())))

            inc.lpad.rerun_fw(fwid)

            clusters_to_launch.append(to_cluster.name)
        self.launch_rockets(clusters_to_launch)
        self.fwid_list = []
        self.load_list()
        self.load_table()

    def launch_rockets(self, clusters_to_launch : typ.List[str])-> None:
        if 'sherlock2' in clusters_to_launch:
            if not get_cluster() == 'sherlock':
                ssh_command = 'ssh -t {}@login.sherlock.stanford.edu'.format(SHERLOCK2_USERNAME)
                command =['bash','-c',"'$CATALOG_LOC/fireworks/launcher.sh'"]
                p = subprocess.Popen(ssh_command.split()+command,stdout=subprocess.PIPE)
                out, _ = p.communicate()
                logging.info(out.decode())
            else:
                pass
        if 'suncat' in clusters_to_launch:
            ssh_command = 'ssh {}@suncatls1.slac.stanford.edu'.format(SUNCAT_USERNAME)
            command =['bash','-c',"'$CATALOG_LOC/fireworks/launcher.sh'"]
            p = subprocess.Popen(ssh_command.split()+command,stdout=subprocess.PIPE)
            out, _ = p.communicate()
            logging.info(out.decode())

    def finish(self)->None:
        self.parent.central_tabWidget.setCurrentIndex(0)
        self.parent.central_tabWidget.removeTab(1)
        self.accept()



#Typing imports
import typing as typ

#External imports
from PyQt5 import QtWebEngineWidgets, QtGui, QtWidgets, QtCore # type: ignore
import ase                                                     # type: ignore
from ase.visualize import view                                 # type: ignore
import os
#Internal Imports
from catalog.misc.plotly_atoms import PlotlyAtoms

class PlotlyView(QtWidgets.QWidget):
    def __init__(self
                ,list_of_atoms_obj : typ.List[ase.Atoms]
                ,parent            : typ.Any            = None
                )-> None:
        super(PlotlyView,self).__init__(parent)
        self.setupUi()
        self.list_of_atoms_obj = list_of_atoms_obj
        self.plotly_atoms      = [PlotlyAtoms(atoms_obj = atoms) for atoms in list_of_atoms_obj]
        self.plotly_htmls      = [plotly_atoms.plot(offline = True) for plotly_atoms in self.plotly_atoms]
        self.current_index     = 0
        self._build_html()
        self._connect_signals()

    def _connect_signals(self) -> None:
        self.previous_button.clicked.connect(self._previous)
        self.next_button.clicked.connect(self._next)
        self.view_button.clicked.connect(self._view_ase)

    def setupUi(self) -> None:
        self.layout          = QtWidgets.QVBoxLayout(self)
        self.viewer          = QtWebEngineWidgets.QWebEngineView()
        self.layout.addWidget(self.viewer)
        self.button_box      = QtWidgets.QGroupBox()
        self.button_layout   = QtWidgets.QHBoxLayout(self.button_box)
        self.previous_button = QtWidgets.QPushButton("Previous Atoms Object")
        self.view_button     = QtWidgets.QPushButton("View in ASE GUI")
        self.next_button     = QtWidgets.QPushButton("Next Atoms Object")
        self.button_layout.addWidget(self.previous_button)
        self.button_layout.addWidget(self.view_button)
        self.button_layout.addWidget(self.next_button)
        self.layout.addWidget(self.button_box)
        self.layout.addStretch(0)

    def _next(self) -> None:
        self.current_index += 1
        self.current_index = min(len(self.plotly_atoms)-1,self.current_index)
        self._build_html()

    def _view_ase(self) -> None:
        view(self.list_of_atoms_obj[self.current_index])

    def _previous(self) -> None:
        self.current_index -= 1
        self.current_index = max(0,self.current_index)
        self._build_html()


    def _build_html(self) -> None:
        raw_html    = '<html><head><meta charset ="utf-8" />'
        raw_html   += '<script src="https://cdn.plot.ly/plotly-latest.min.js"></script></head>'
        raw_html   += '<body>'
        raw_html   += '<div align="center">'
        raw_html   += self.plotly_htmls[self.current_index]
        raw_html   += '</div>'
        raw_html   += '</body></html>'
        self.viewer.setHtml(raw_html)
        self.viewer.page()
        # self.viewer.update()



#External imports
import os

#Internal Imports
"""
This file contains the enviromental variables and job submission settings
"""

###################
#AUTO-GENERATED KEYS
###################


###################
#USER SPECIFIED KEYS
###################
"""
Current required enviromental variables:
    - USER
        The name the user wants to use as default login to new clusters
    - HOSTNAME
        The name of the machine used to identify the cluster please use 'local'
        for non-job running machines
    - SHERLOCK2_USERNAME
        Login name for sherlock2
    - SUNCAT_USERNAME
        Login name for suncat,suncat2, and suncat3
    - LAUNCHPAD_YAML
        path to the launchpad yaml to load the fireworks launchpad
"""
CATALOG_LOC = os.getenv('CATALOG_LOC')
assert not CATALOG_LOC == None, 'MISSING ENVIROMENTAL VARIABLE: CATALOG_LOC'

USER = os.getenv('USER')
assert not USER == None, 'MISSING ENVIROMENTAL VARIABLE: USER'

HOSTNAME     = os.getenv('HOSTNAME')
assert not HOSTNAME == None, 'MISSING ENVIROMENTAL VARIABLE: HOSTNAME'

#Sherlock2 Keys
###############
SHERLOCK2_USERNAME = os.getenv('SHERLOCK2_USERNAME')
assert not SHERLOCK2_USERNAME == None, 'MISSING ENVIROMENTAL VARIABLE: SHERLOCK2_USERNAME'

#Suncat Keys
###############
SUNCAT_USERNAME = os.getenv('SUNCAT_USERNAME')
assert not SUNCAT_USERNAME == None, 'MISSING ENVIROMENTAL VARIABLE: SUNCAT_USERNAME'

#Fireworks keys
###############
LAUNCHPAD_YAML   = os.path.join(CATALOG_LOC,'fireworks','FW_CONFIG_DIR','my_launchpad.yaml')
assert not LAUNCHPAD_YAML == None, 'MISSING ENVIROMENTAL VARIABLE: LAUNCHPAD_YAML'


# ######################
# # Job Submission Settings
# # --------------------

def allocate(job,n):
    """
    Take in a Job object and an integer, n. Return a list of n Cluster objects.
    """
    return ['sherlock2']

def guessTime(job) -> int:
    return 36

def guessNodes(job) -> int:
    return 2



#Typing imports
import typing as typ

#External imports
from fireworks.queue.queue_launcher import rapidfire # type: ignore
import os

#Internal Imports
from catalog.fw.incomplete import lpad
from catalog.jobs.cluster import Cluster, cluster_dict
from catalog.catalog_config import HOSTNAME

def launch(clust_str : str = get_cluster()):
    cluster = cluster_dict[clust_str]
    rapidfire(lpad, cluster.fworker, cluster.qadapter, launch_dir=os.path.join(CATALOG_LOC,'fireworks','FW_CONFIG_DIR'))

if __name__ == '__main__':
    launch(HOSTNAME)



import argparse
import sys

from catalog import __version__
from importlib import import_module # type: ignore


commands = [
    ('gui', 'catalog.cli.gui')]

def main(prog ='catalog', description='catalog command line tool',
         version=__version__, commands=commands, hook=None, args=None):
    parser = argparse.ArgumentParser(prog=prog, description=description)
    parser.add_argument('--version', action='version',
                        version='%(prog)s-{}'.format(version))
    parser.add_argument('-T', '--traceback', action='store_true')
    subparsers = parser.add_subparsers(title='Sub-commands',
                                       dest='command')

    subparser = subparsers.add_parser('help',
                                      description='Help',
                                      help='Help for sub-command')
    subparser.add_argument('helpcommand', nargs='?')

    functions = {}
    parsers = {}
    for command, module_name in commands:
        cmd = import_module(module_name).CLICommand
        subparser = subparsers.add_parser(
            command,
            help=cmd.short_description,
            description=getattr(cmd, 'description', cmd.short_description))
        cmd.add_arguments(subparser)
        functions[command] = cmd.run
        parsers[command] = subparser

    if hook:
        args = hook(parser, args)
    else:
        args = parser.parse_args(args)

    if args.command == 'help':
        if args.helpcommand is None:
            parser.print_help()
        else:
            parsers[args.helpcommand].print_help()
    elif args.command is None:
        parser.print_usage()
    else:
        f = functions[args.command]
        try:
            if f.__code__.co_argcount == 1:
                f(args)
            else:
                f(args, parsers[args.command])
        except KeyboardInterrupt:
            pass
        except Exception as x:
            if args.traceback:
                raise
            else:
                l1 = '{}: {}\n'.format(x.__class__.__name__, x)
                l2 = ('To get a full traceback, use: {} -T {} ...'
                      .format(prog, args.command))
                parser.error(l1 + l2)



import setuptools, re # type: ignore

with open('catalog/__init__.py') as fd:
    version = re.search("__version__ = '(.*)'", fd.read()).group(1)

setuptools.setup(
    name="CC-CataLog"
    ,version=version
    ,url="https://github.com/statt8900/CataLog"
    ,author="Michael Statt, Kris Brown"
    ,author_email="mstatt@stanford.edu, ksb@stanford.edu"
    ,description="Tools for computational catalyst research"
    ,license='APACHE LICENSE, VERSION 2.0'
    ,packages=['catalog'
             ,'catalog.data'
             ,'catalog.datalog'
             ,'catalog.fw'
             ,'catalog.gendata'
             ,'catalog.gui'
             ,'catalog.jobs'
             ,'catalog.misc'
             ,'catalog.cli'
    ]
    ,entry_points={'console_scripts': ['catalog=catalog.cli.main:main']}
    ,install_requires=['ase==3.16'
                      ,'numpy'
                      ,'pymysql'
                      ,'sqlparse'
                      ,'python-sql'
                      ,'pymatgen'
                      ,'networkx'
                      ,'plotly'
                      ,'fireworks'
                      ,'mysqlclient'
                      ,'PyQt5'
                      ,'qdarkstyle'
                      ,'prettytable'
                      ,'CC-Adsorber'
    ]
    ,python_requires='>3.6, <4'
    ,classifiers=[
        'Development Status :: 4 - Beta'
       ,'Intended Audience :: Developers'
       ,'Topic :: Scientific/Engineering :: Chemistry'
       ,'Programming Language :: Python :: 3.6'
    ]
)



#Typing imports
import typing as typ
if typ.TYPE_CHECKING:
    from catalog.jobs.jobs import Job

#External imports
import subprocess,os
#--fireworks imports
from fireworks.core.fworker import FWorker                                     # type: ignore
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter # type: ignore

#Internal Imports
from catalog.catalog_config import USER, SUNCAT_USERNAME

"""
The Cluster object and its only instances are declared

NEED TO ADD:
    - SHERLOCK2
    - CORI
    - EDISON
"""
################################################################################

class Cluster(object):
    """
    Helpful docstring
    """
    def __init__(self
                ,name      : str
                ,fworker   : str
                ,launchdir : str
                ,qfunc     : typ.Callable[[int,int,int]
                                         ,typ.Dict[str,typ.Any]]
                ,qtype     : str
                ) -> None:
        self.name       = name
        self.fworker    = FWorker(fworker,query = {})
        self.qfunc      = qfunc
        self.qadapter   = CommonAdapter(qtype)
        self.launchdir  = launchdir

    def __repr__(self)->str: return self.name


def printTime(floatHours : float) -> str:
    """
    docstring
    """
    intHours = int(floatHours)
    return "%02d:%02d" % (intHours,(floatHours-intHours)*60)

###########################################################################

def sherlock2Q(timeInHours  : int = 1
              ,nodes        : int = 1
              ,cores        : int = 16
              ) -> typ.Dict[str,typ.Any]:


    return {'_fw_name'       : 'CommonAdapter'
           ,'_fw_q_type'     : 'SLURM'
           ,'rocket_launch'  : 'rlaunch -c $CATALOG_LOC/FW_CONFIG_DIR singleshot'
           ,'nodes'          : nodes
           ,'ntasks_per_core': 1
           ,'mem'            : 64000
           ,'walltime'       : printTime(timeInHours)+':00'
           ,'queue'          : 'iric'
           ,'pre_rocket'     : 'source $CATALOG_LOC/.env/bin/activate'
           }

def sherlock2QGPAW(timeInHours : int = 1
                  ,nodes       : int = 1
                  ,cores       : int = 20
                  ) -> typ.Dict[str,typ.Any]:
    if cores is not None:
        assert nodes==1 and cores < 17
    else:
        cores = 16
    return {'_fw_name':         'CommonAdapter'
            ,'_fw_q_type':      'SLURM'
            ,'rocket_launch':   'rlaunch -c $CATALOG_LOC/FW_CONFIG_DIR singleshot'
            ,'nodes':           nodes
            ,'ntasks_per_core': 1
            ,'walltime':        printTime(timeInHours)+':00'
            ,'queue':           'iric'
            ,'pre_rocket':      'export PATH=$PATH:/scratch/users/{0}/fireworks/fireworks_env/bin/;source /scratch/PI/suncat/sw/env.bash;export OMP_NUM_THREADS=1;export PYTHONPATH=/scratch/users/{0}/gpaw/gpaw_sg15/lib/python2.7/site-packages:$PYTHONPATH;export PATH=/scratch/users/{0}/gpaw/gpaw_sg15/bin:$PATH;export GPAW_SETUP_PATH=/scratch/users/{0}/gpaw/gpaw_sg15/norm_conserving_setups'.format(USER)
            }

###########################################################################
def suncatQ(timeInHours : int               = 1
           ,nodes       : int               = 1
           ,cores       : typ.Optional[int] = None
           ) -> typ.Dict[str,typ.Any]:

    if cores is not None:
        assert nodes==1 and cores < 9
        ntasks = cores
    else:
        ntasks = 8*nodes

    return {'_fw_name':        'CommonAdapter'
            ,'_fw_q_type':     'LoadSharingFacility'
            ,'rocket_launch':  'rlaunch -c $CATALOG_LOC/FW_CONFIG_DIR singleshot'
            ,'ntasks':         ntasks
            ,'walltime':       printTime(timeInHours)
            ,'queue':          'suncat'
            ,'pre_rocket':     'source $CATALOG_LOC/.env/bin/activate'
            }
###########################################################################
def suncat2Q(timeInHours : int
            ,nodes       : int
            ,cores       : typ.Optional[int] = None
            ) -> typ.Dict[str,typ.Any]:

    if cores is not None:
        assert nodes==1 and cores < 13
        ntasks = cores
    else:
        ntasks = 12*nodes

    return {'_fw_name':       'CommonAdapter'
            ,'_fw_q_type':    'LoadSharingFacility'
            ,'rocket_launch': 'rlaunch -c $CATALOG_LOC/FW_CONFIG_DIR singleshot'
            ,'ntasks':        ntasks
            ,'walltime':      printTime(timeInHours)
            ,'queue':         'suncat2'
            ,'pre_rocket':    'source /nfs/slac/g/suncatfs/sw/py2.7.13/env.bash'
            }

def coriQ(timeInHours : int = 1
         ,nodes       : int = 1
         ,cores       : int = 32
         ) -> typ.Dict[str,typ.Any]:

    return {'_fw_name':         'CommonAdapter'
            ,'_fw_q_type':      'SLURM'
            ,'rocket_launch':   'cd /global/cscratch1/sd/krisb/fireworks;rlaunch singleshot'.format(USER)
            ,'nodes':           nodes
            ,'ntasks_per_node': cores
            ,'walltime':        printTime(timeInHours)+':00'
            ,'queue':           'regular'
            ,'qos':             'normal'
            ,'pre_rocket':      ''
            ,'constraint':      'haswell'
            ,'logdir':          '/global/cscratch1/sd/krisb/fireworks/logs/'}

def coriDBQ(timeInHours : int = 1
           ,nodes       : int = 1
           ,cores       : int =32
           ) -> typ.Dict[str,typ.Any]:

    return {'_fw_name':         'CommonAdapter'
            ,'_fw_q_type':      'SLURM'
            ,'rocket_launch':   'cd /global/cscratch1/sd/krisb/fireworks;rlaunch singleshot'.format(USER)
            ,'nodes':           nodes
            ,'ntasks_per_node': cores
            ,'walltime':        printTime(min(0.5,timeInHours))+':00'
            ,'queue':           'debug'
            ,'qos':             'normal'
            ,'pre_rocket':      ''
            ,'constraint':      'haswell'
            ,'logdir':          '/global/cscratch1/sd/krisb/fireworks/logs/'}

def ediDBQ(timeInHours : int = 1
          ,nodes       : int = 1
          ,cores       : int = 32
          ) -> typ.Dict[str,typ.Any]:

    return {'_fw_name':         'CommonAdapter'
            ,'_fw_q_type':      'SLURM'
            ,'rocket_launch':   'cd /scratch1/scratchdirs/krisb/fireworks;rlaunch singleshot'.format(USER)
            ,'nodes':           nodes
            ,'ntasks_per_node': cores
            ,'walltime':        printTime(min(0.5,timeInHours))+':00'
            ,'queue':           'debug'
            ,'qos':             'normal'
            ,'pre_rocket':      ''
            ,'logdir':          '/scratch1/scratchdirs/krisb/fireworks/logs/'}

###########################################################################
#Get the location of the CataLog folder on each of the clusters to find the launchdirs
#All jobs get launched in the $CATALOG_LOC/jobs folder
sherlock2_CATALOG_LOC = subprocess.check_output(['ssh', '-XY','{}@login.sherlock.stanford.edu'.format(USER),'echo',"$CATALOG_LOC"]).decode().strip('\n')
sherlock2_launchdir   = os.path.join(sherlock2_CATALOG_LOC,'fireworks','jobs')

suncat_CATALOG_LOC    = subprocess.check_output(['ssh', '-XY','{}@suncatls1.slac.stanford.edu'.format(USER),'echo',"$CATALOG_LOC"]).decode().strip('\n')
suncat_launchdir      = os.path.join(suncat_CATALOG_LOC,'fireworks','jobs')
###########################################################################
###########################################################################


sherlock2        = Cluster('sherlock2',  'sherlock2',   sherlock2_launchdir,       sherlock2Q,'SLURM')
sherlock2gpaw    = Cluster('sherlock2',  'sherlock2',   sherlock2_launchdir,       sherlock2QGPAW,'SLURM')


suncat          = Cluster('suncat',     'suncat',      suncat_launchdir, suncatQ, 'LoadSharingFacility')
suncat2         = Cluster('suncat2',    'suncat',      suncat_launchdir, suncat2Q,'LoadSharingFacility')

cori            = Cluster('cori',       'cori',        '/global/cscratch1/sd/krisb/fireworks/jobs/',            coriQ,'SLURM')
cori_debug      = Cluster('cori-debug', 'cori',        '/global/cscratch1/sd/krisb/fireworks/jobs/',            coriDBQ,'SLURM')
edi_debug       = Cluster('edi-debug',  'edison',      '/scratch1/scratchdirs/krisb/fireworks/jobs/',           ediDBQ,'SLURM')

cluster_dict = {x.name:x for x in [sherlock2,suncat, suncat2]}

def str2Cluster(job : typ.Any
               ,s : str
               ,n : int
               ) -> typ.List[Cluster]:
    assignDict = {'sherlock2': {'gpaw':sherlock2gpaw
                               ,'quantumespresso':sherlock2}
                    ,'suncat': {'quantumespresso':suncat}
                    ,'suncat2': {'quantumespresso':suncat2}
                 ,'cori':
                    {'vasp':cori_debug}#cori
                 ,'edison':
                    {'vasp':edi_debug}
                 }

    return [assignDict[s.lower()][job.params['dftcode']]]*n

###################
#Archived Scripts
###################
# ###########################################################################
#
# def sherlockQ(timeInHours : int = 1
#              ,nodes       : int = 1
#              ,cores       : int = 16
#              ) -> typ.Dict[str,typ.Any]:
#     if cores is not None:
#         assert nodes==1 and cores < 17
#     else:
#         cores = 16
#     return {'_fw_name':         'CommonAdapter'
#             ,'_fw_q_type':      'SLURM'
#             ,'rocket_launch':   'cd /scratch/users/{0}/fireworks;rlaunch singleshot'.format(user)
#             ,'nodes':           nodes
#             ,'ntasks_per_core': 1
#             ,'mem'            : 64000
#             ,'walltime':        printTime(timeInHours)+':00'
#             ,'queue':           'iric,owners'
#             ,'pre_rocket':      'source {0}/.env/bin/activate'.format(CataLogPath)#'source /scratch/PI/suncat/sw/env.bash' #
#             ,'logdir':          '/scratch/users/{0}/fireworks/logs/'.format(user)}
#
# def sherlockQGPAW(timeInHours : int = 1
#                  ,nodes       : int = 1
#                  ,cores       : int = 16
#                  ) -> typ.Dict[str,typ.Any]:
#     if cores is not None:
#         assert nodes==1 and cores < 17
#     else:
#         cores = 16
#     return {'_fw_name':         'CommonAdapter'
#             ,'_fw_q_type':      'SLURM'
#             ,'rocket_launch':   'cd /scratch/users/{0}/fireworks;rlaunch singleshot'.format(user)
#             ,'nodes':           nodes
#             ,'ntasks_per_node': cores
#             ,'walltime':        printTime(timeInHours)+':00'
#             ,'queue':           'iric'
#             ,'qos':             'iric'
#             ,'pre_rocket':      ';'.join(["source /scratch/PI/suncat/sw/env.bash"
#                                         ,"source ~/scripts/rc/RCsher.sh"
#                                         ,"export OMP_NUM_THREADS=1"
#                                         ,"export PYTHONPATH=/scratch/users/{0}/gpaw/ggafirst/install/lib/python2.7/site-packages:/scratch/users/{0}/gpaw/gpaw_sg15/lib/python2.7/site-packages:$PYTHONPATH".format(user)
#                                         ,"export PATH=/scratch/users/{0}/gpaw/ggafirst/install/bin:$PATH".format(user)
#                                         ,"export GPAW_SETUP_PATH=/scratch/users/{0}/gpaw/gpaw_sg15/norm_conserving_setups".format(user)])
#             ,'logdir':             '/scratch/users/{0}/fireworks/logs/'.format(user)}
#
###########################################################################



# Typing Modules
import typing as typ

if typ.TYPE_CHECKING:
    #Type Checking Imports
    import sql                                  # type: ignore
    from catalog.jobs.jobs import Job

# External Modules
import warnings
warnings.filterwarnings("ignore", message="Moved to ase.neighborlist")
warnings.filterwarnings("ignore", message="matplotlib.pyplot")

import json,pickle,copy,inspect,itertools

import ase                                  # type: ignore
import ase.visualize as viz                 # type: ignore
from ase.build import bulk                  # type: ignore
import numpy as np                          # type: ignore
from copy import deepcopy                   # type: ignore

# Internal modules
import catalog.data.emt  as emt

import catalog.jobs.jobs                 as jobs
import catalog.gendata.surfFuncs         as surfFuncs
import catalog.gendata.interstitialFuncs as interstitialFuncs
import catalog.datalog.db_utils          as db
import catalog.fw.incomplete             as manage

from catalog.misc.sql_utils   import *
from catalog.misc.utilities   import launch,merge_dicts,flatten
from catalog.misc.print_parse import abbreviate_dict,read_on_sher
from catalog.misc.atoms       import (make_atoms,nonmetal_symbs,cell_to_param
                                      ,remove_atom,remove_atoms, fix_negative_unit_cells
                                      ,get_sites,traj_to_json,json_to_traj,collision,reorient_z_ase)
# from catalog.gendata.ads_site import AdsorbateSiteVector

from catalog.jobs.repeat      import RepeatChecker
from catalog.jobs.versions    import current_version

#Type Aliases
# filter_func_type    = typ.Callable[[typ.List['AdsorbateSiteVector']],typ.List['AdsorbateSiteVector']]
transform_func_type = typ.Callable[[dict],typ.List[dict]]

################################################################################
"""
Functions to create new jobs based on existing jobs in the database

varies
modify
bulk_modulus
vibs
bare
adsorb
interstitial
"""

dbp     = '/scratch/users/ksb/share/suncatdata.db' #default DataBasePath
repeats = RepeatChecker() # Initialize Repeat Checker when loading script

def calc_args(drop          : typ.List[str]             = []
             ,connectinfo   : ConnectInfo               = db.realDB
             ) -> typ.List[sql.Column]:
    """
    Returns list of SQL Column objects that can be used in a query
    """
    drop.append('calc_id') # always drop this one
    drop.append('calc_other_id') # always drop this one
    drop += ['kx','ky','kz'] # always drop this one
    calc_cols       = [getattr(tbl_calc,col) for col in
                db.Query(table=tbl_calc,connectinfo=connectinfo).col_names()
                if col not in drop]
    calc_other_cols = [getattr(tbl_calc_other,col) for col in
                db.Query(table=tbl_calc_other,connectinfo=connectinfo).col_names()
                if col not in drop]
    return calc_cols + calc_other_cols


#############
# Check_funcs
#-------------

def check_params(j : 'Job') -> bool:
    q = '\n\nDo the following params look right?\n\n%s\n\n(y/n)--> '%(abbreviate_dict(j.params))
    return 'y' in input(q).lower()

def check_atoms(j  : 'Job') -> bool:
    viz.view(j.atoms)
    q = 'Do the structure/params look right?\n%s\n(y/n)--> '%(abbreviate_dict(j.params))
    return 'y' in input(q).lower()
#################
# CREATE NEW JOBS
#################


# NON-gendata methods (jobs that must be built from scratch, not from DB)
def latticeopt(name                 : str
              ,structure            : str
              ,dftcode              : str
              ,pw                   : int
              ,xc                   : str
              ,psp                  : str
              ,walltime             : typ.Optional[float]   = None
              ,nodes                : typ.Optional[int]     = None
              ,clust_str            : typ.Optional[str]     = None # default submission if None
              ,kpts                 : typ.List[int]         = [8,8,8]
              ,econv                : float                 = 5e-4
              ,xtol                 : float                 = 0.005  # pretty relevant parameters
              ,fmax                 : float                 = 0.05
              ,mixing               : float                 = 0.1
              ,nmix                 : int                   = 10
              ,maxstep              : int                   = 250
              ,nbands               : int                   = -12
              ,sigma                : float                 = 0.1 # other parameters
              ,gga                  : str                   = 'PE'
              ,gamma                : int                   = 1
              ,ismear               : int                   = 0
              ,algo                 : str                   = 'normal'
              ,prec                 : str                   = 'normal'
              ,mixingtype           : str                   = 'plain'
              ,ibrion               : int                   = 2
              ,nelmdl               : int                   = -9
              ,lreal                : str                   = 'auto'
              ,lvhar                : int                   = 1
              ) -> None:
    """
    Perform lattice optization on a metal element.
    Optionally provide an associated list of structures if non-equilibrium
    structures are wanted.
    """
    assert structure in ['fcc','bcc','hcp','fcc-unit','bcc-unit','zincblende','zincblende-unit','cscl','rocksalt','cscl-unit','rocksalt-unit','triclinic']
    unit  = 'unit' not in structure
    cubic = 'cubic' in structure # not possible but maybe in future?

    crystalstructure = structure # Apply transformations to make ASE compatible?

    itraj = bulk(name, crystalstructure= crystalstructure, orthorhombic=unit, cubic=cubic)

    params = {'inittraj':traj_to_json(itraj),'structure':structure
            ,'dftcode':dftcode,'pw':pw,'xc':xc,'psp':psp,'kpts':kpts
            ,'econv':econv,'xtol':xtol,'fmax':fmax,'maxstep':maxstep
            ,'sigma':sigma,'jobkind':'latticeopt'}

    qe_dict   = {'nbands':nbands,'mixing':mixing,'nmix':nmix,'mixingtype':mixingtype}
    gpaw_dict = {'nbands':nbands,'mixing':mixing,'nmix':nmix}
    vasp_dict = {'gga':gga,'luse_vdw': 1 if xc == 'BEEF' else 0
                ,'zab_vdw':      -1.8867 if xc == 'BEEF' else None
                ,'gamma':gamma,'ismear':ismear,'algo':algo
                ,'lreal':lreal,'lvhar':lvhar,'prec':prec
                ,'ionic_steps':50 ,'ibrion':ibrion,'nelmdl':nelmdl}

    meta_dict = {'quantumespresso':qe_dict,'gpaw':gpaw_dict,'vasp':vasp_dict}
    input_dict = merge_dicts([params,meta_dict[dftcode]])       # type: ignore

    j = jobs.Job(input_dict)
    if not repeats.check_repeat(j):
        msg =  'New job, sure you want to submit the following? '+abbreviate_dict(input_dict)
        if 'y' in input(msg).lower():
            j.submit(repeats, walltime = walltime, clust_str = clust_str,nodes = nodes)
        else:
            print('No job submitted')
    else:
        print('This job is not new')

# Gendata-creating functions (must call .submit() on the output)
def varies(field            : str
          ,rang             : typ.List[typ.Any]
          ,constraint       : typ.List[typ.Any]     = []
          ,limit            : typ.Optional[int]     = None
          ,check_func       : typ.Callable          = check_atoms
          ) -> 'GenData':
    """
    Create multiple jobs per old job by varying 'field' over range 'rang'
    """
    if field == 'delta': default_constraint = VIB_ # What other cases is it ok to vary a parameter for a non-relax/latticeopt job
    else: default_constraint = OR(LATTICEOPT_,VCRELAX_,AND(RELAX_,MOLECULE_))

    cnst = AND(default_constraint,*constraint)

    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['kpts'])+
                   [As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
                    ,As(JOBTYPE,'jobkind'),As(KPTDEN_X/2+KPTDEN_Y/2,'kptden')]
                    ,cnst,table=tbl_relaxfinal,group=JOB
                    ,limit=limit)

    return GenData(query,check_func = check_atoms
                  ,transform=[preserve_dwrat
                            ,job_name_kwargs
                            ,vary_dict(field,rang)])

def modify(modifying_dict   : typ.Dict[typ.Any,typ.Any]
          ,constraint       : typ.List[typ.Any]         = []
          ,limit            : typ.Optional[int]         = None
          ,check_func       : typ.Callable              = check_atoms
          )->'GenData':
    """
    Create 1 new job per old job (specified by constraint) by
        modifying params with a dictionary of unary functions.
    K-Point density is preserved
    """
    default_constraint = OR(LATTICEOPT_,VCRELAX_,AND(RELAX_,MOLECULE_))
    cnst               = AND(default_constraint,*constraint)

    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['kpts'])
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
                        ,As(JOBTYPE,'jobkind')
                        ,As(KPTDEN_X/2+KPTDEN_Y/2,'kptden')]
                    ,cnst,group=JOB,table=tbl_relaxfinal_all,limit=limit)

    return GenData(query,check_func = check_atoms
                  ,transform=[print_params()
                            ,preserve_dwrat
                            ,job_name_kwargs
                            ,update_keys(modifying_dict)])


def xc_contribs(constraints          :typ.List[typ.Any]       = []
               ,limit               :int                     = 1
               ,check_func       : typ.Callable              = check_atoms
               ) -> 'GenData':
    """Get the XC contributions"""
    default_constraint = AND(LATTICEOPT_,PW_(1500),KSB_)    # type: ignore
    cnst               = AND(default_constraint,*constraints)

    query=db.Query([As(RAW,'inittraj')]+calc_args()
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')]
                    ,cnst,group=JOB,table=tbl_relaxfinal_all
                    ,limit=limit)

    return GenData(query #,check_func = check_params
                  ,transform=[insert_keys(jobkind='xc')
                            ,update_keys({'job_name':lambda x: x+'_xc_contribs'})
                            ])


def bulk_modulus(constraints         :typ.List[typ.Any]
                ,limit              :typ.Optional[int]      = None
                ,check_func       : typ.Callable              = check_params
                ) -> 'GenData':
    """
    Calculate bulk modulus
    """
    default_constraint = LATTICEOPT_
    cnst               = AND(default_constraint,*constraints)

    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['kpts'])
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
                    ,As(JOBTYPE,'jobkind'),As(KPTDEN_X/2+KPTDEN_Y/2,'kptden')]
                    ,cnst,group=JOB,table=tbl_relaxfinal_all,limit=limit)

    return GenData(query,check_func = check_func
                  ,transform=[insert_keys(jobkind='bulkmod',strain=0.03)
                            ,update_keys({'job_name':lambda x: x+'_bulkmod'})
                            ,job_name_kwargs])

def vibs(constraints                 :typ.List[typ.Any]
        ,limit                      :typ.Optional[int]      = None
        ,check_func       : typ.Callable              = check_atoms
        ) -> 'GenData':
    """
    Calculate vibrational modes
    """
    default_constraint = RAWENG!=None
    cnst               = AND(default_constraint,*constraints)
    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['xtol','cell_dofree','cell_factor'])
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent'),FACET_CATALOG,ADSORBATES
                    ,As(JOBTYPE,'jobkind')],cnst,table=tbl_relaxfinal_all
                    ,group=JOB,limit=limit)

    return GenData(query,check_func = check_func
                  ,transform=[insert_keys(jobkind='vib',delta=0.04)
                            ,update_keys({'job_name':lambda x: x+'_vib'})
                            ,vibrate_nonmetals
                            ,job_name_kwargs])

def repeat_traj(constraints       : typ.List[typ.Any]
               ,repeat            : typ.Tuple[int,int,int]
               ,fix_cell_problems : bool = True
               ,adjust_kpts       : bool = True
               ,limit             : typ.Optional[int] = None
               ,verbose           : bool = True
               ,check_func       : typ.Callable              = check_atoms
               ) -> 'GenData':
    default_constraint     = AND(RELAXORLAT_,SURFACE_)
    cnst                   = AND(default_constraint,*constraints)

    q_args = ( [As(RAW,'inittraj')]
             + calc_args(drop=['xtol'])
             + [As(STRUCTURE,'structure'),JOBNAME,As(ADSORBATES,'adsorbates'),As(STORDIR,'parent')
               ,As(JOBTYPE,'jobkind'),As(FACET_CATALOG,'facet')])

    query=db.Query(cols=q_args,constraints=cnst,group=JOB
                    ,table=tbl_relaxfinal_all,limit=limit, verbose = True)

    return GenData(query,check_func = check_func
                  ,transform=[repeat_atoms(repeat            = repeat
                                          ,fix_cell_problems = fix_cell_problems
                                          ,adjust_kpts       = adjust_kpts)
                            ,job_name_kwargs]
                  , verbose= verbose)

def bare_all_facets(constraints : typ.List[typ.Any]
                   ,xy          : typ.Tuple[int,int]
                   ,layers      : int
                   ,symmetric   : bool
                   ,vacuum      : float
                   ,kpts        : typ.List[int]
                   ,max_index   : int = 1
                   ,constrained : typ.Optional[typ.List[int]] = None
                   ,limit       : typ.Optional[int] = None
                   ,check_func  : typ.Callable[[typ.Any],bool] = check_atoms
                   ,verbose     : bool = True
                   ) -> 'GenData':
    """Generate all facets of a bare surface with a max miller index equal to max_index"""
    default_constraint = AND(RAWENG!=None,BULK_)
    cnst               = AND(default_constraint,*constraints)

    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['xtol','kpts','kptden','dipole','cell_dofree','cell_factor'])
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
                        ,As(JOBTYPE,'jobkind')]
                    ,cnst,group=JOB,table=tbl_relaxfinal_all,limit=limit)
    if constrained is None:
        constrained = list(range(int(layers/2+1),layers+1))

    return GenData(query,check_func = check_func
                  ,transform=[insert_keys(jobkind='relax',adsorbates='[]')
                             ,print_params()
                             ,insert_keys(kpts=kpts)
                             ,make_bare_all_facets(max_index,xy,layers,constrained,symmetric,vacuum)
                             ,job_name_kwargs
                             ,print_params()], verbose= verbose)

def bare(constraints : typ.List[typ.Any]
        ,xy          : typ.Tuple[int,int]
        ,layers      : int
        ,symmetric   : bool
        ,vacuum      : float
        ,kpts        : typ.List[int]
        ,facet       : typ.Tuple[int,int,int]
        ,constrained : typ.Optional[typ.List[int]] = None
        ,limit       : typ.Optional[int] = None
        ,verbose     : bool = True
        ,check_func  : typ.Callable = check_atoms
        ) -> 'GenData':
    """Generate a bare surface"""
    default_constraint = AND(RAWENG!=None,BULK_)
    cnst               = AND(default_constraint,*constraints)

    query=db.Query([As(RAW,'inittraj')]+calc_args(drop=['xtol','kpts','kptden','dipole','cell_dofree','cell_factor'])
                    +[As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
                        ,As(JOBTYPE,'jobkind')]
                    ,cnst,group=JOB,table=tbl_relaxfinal_all,limit=limit)
    if constrained is None:
        constrained = list(range(int(layers/2+1),layers+1))

    return GenData(query,check_func = check_func
                  ,transform=[insert_keys(jobkind='relax',adsorbates='[]',facet=json.dumps(facet))
                             ,print_params()
                             ,insert_keys(kpts=kpts)
                             ,update_keys({'job_name':bare_name(facet,xy,layers)})
                             ,make_bare_slab(facet,xy,layers,constrained,symmetric,vacuum)
                             ,job_name_kwargs
                             ,print_params()], verbose= verbose)


def adsorb(constraints          : typ.List[typ.Any] = []
          ,adsname              : str               = 'H'
          ,ads_filter_list      : list              = []
          ,slab_constraint_dict : dict              = {}
          ,ads_constraint_dict  : dict              = {}
          ,height               : float             = 2.0
          ,limit                : typ.Optional[int] = None
          ,verbose              : bool              = True
          ,check_func       : typ.Callable              = check_atoms
          ) -> 'GenData':
    """
    adsfilterlist is a list of functions [(Ads,Site,Vector)] -> [(Ads,Site,Vector)]
    """
    from catalog.misc.sql_utils import AND
    from adsorber import filter_functions as ff
    from adsorber.objects.adsorbate import get_ads

    default_constraint     = AND(RELAXORLAT_,SURFACE_)
    cnst                   = AND(default_constraint,*constraints)
    ads_list = list(map(get_ads,adsname.split())) # Adsorbate class

    q_args = ( [As(RAW,'inittraj')]
             + calc_args(drop=['xtol','cell_dofree','cell_factor'])
             + [As(STRUCTURE,'structure'),JOBNAME,As(ADSORBATES,'adsorbates'),As(STORDIR,'parent')
               ,As(JOBTYPE,'jobkind'),As(FACET_CATALOG,'facet')])

    query=db.Query(cols=q_args,constraints=cnst,group=JOB
                    ,table=tbl_relaxfinal_all,limit=limit, verbose = verbose)

    return GenData(query,check_func = check_func
                  ,transform=[constrain_slab(slab_constraint_dict)
                            ,add_adsorbates(ads_list,ads_filter_list, constraint_dict = ads_constraint_dict,height = height)
                            ,job_name_kwargs], verbose = verbose)


def vacancy(constraints : typ.List[typ.Any] = []
           ,limit       : typ.Optional[int] = None
           ,verbose     : bool              = True
           ,check_func       : typ.Callable              = check_atoms
           ) -> 'GenData':
    """
    adsfilterlist is a list of functions [(Ads,Site,Vector)] -> [(Ads,Site,Vector)]
    """
    from catalog.misc.sql_utils import AND
    default_constraint     = AND(RELAXORLAT_)
    cnst                   = AND(default_constraint,*constraints)

    q_args = ( [As(RAW,'inittraj')]
             + calc_args(drop=['xtol','cell_dofree','cell_factor'])
             + [As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
               ,As(JOBTYPE,'jobkind'),As(FACET_CATALOG,'facet')])

    query=db.Query(cols=q_args,constraints=cnst,group=JOB
                    ,table=tbl_relaxfinal_all,limit=limit)

    return GenData(query
                  ,check_func = check_func
                  ,transform=[select_vacancy_index()
                             ,job_name_kwargs]
                  , verbose = verbose)


def fbl(constraint     : typ.List[typ.Any]           = []
       ,bonded_inds    : typ.Optional[typ.List[int]] = None
       ,step_size      : typ.Optional[float]         = None
       ,energy_cut_off : typ.Optional[float]         = None
       ,fix            : typ.Optional[float]         = None
       ,limit          : typ.Optional[int]           = None
       ,check_func     : typ.Callable                = check_params
       ) -> 'GenData':

    if step_size is None: step_size = 0.1
    if energy_cut_off is None: energy_cut_off = 10.
    if fix is None: fix = 0.5

    default_constraint     = AND(RELAX_)
    cnst                   = AND(default_constraint,*constraint)
    q_args = ( [As(RAW,'inittraj')]
             + calc_args(drop=['xtol','cell_dofree','cell_factor'])
             + [As(STRUCTURE,'structure'),JOBNAME,As(STORDIR,'parent')
               ,As(JOBTYPE,'jobkind'),ADSORBATES, As(FACET_CATALOG,'facet')])

    query=db.Query(cols=q_args,constraints=cnst,group=JOB
                    ,table=tbl_relaxfinal_all,limit=limit)

    return GenData(query,check_func = check_params
                  ,transform=[insert_keys(jobkind='fbl',step_size=step_size, energy_cut_off=energy_cut_off, fix = fix)
                            ,select_bonded_inds()
                            ,job_name_kwargs])
#################################################################################

class GenData(object):
    """
    Use an existing database to generate new data by creating Jobs objects
    Its value lies in its only public method: submit
    """
    def __init__(self
                ,query      : db.Query
                ,transform  : typ.List[transform_func_type]
                ,check_func : typ.Callable[[typ.Any],bool]  = lambda x: True
                ,verbose    : bool                          = True
                ) -> None:
        """
        Args:

            constraint :: String
                - Raw SQL to be inserted in the WHERE clause.
                - Determines which subset of existing jobs are used to generate new jobs

            table :: String
                Table used for SQL query
                If one wants to use a job's initial atoms, one join job on initatoms

            limit :: Int
                Restrict number of jobs that meet the query

            transform :: [Dict -> Either Dict [Dict]]
                - Apply transformations to the full row returned from the SQL query
                - Optionally 'branch' by returning a list of dictionaries
                - effects of transformations compose to finally produce Job input

            atoms_to_dicts :: ase.Atoms -> [Dict]
            calc_to_dicts  :: Dict      -> [Dict]
                - These functions are used to modify the previous job
                - If a list greater than length 1 is returned, multiple jobs will be created
                - All combinations of the [Atoms] and [calcDict] will be created
                - To remove a calc parameter (e.g. creating surface slab from latticeopt job)
                    it is enough to do: def c_2_c(xtol,**kwargs): return [kwargs]
            name_to_name :: String -> String
                apply some constant transformation to the previous job_name
            params_to_params :: Dict -> Either Dict [Dict]
                Before the Job is created, apply a final transformation to params
                This is a less restrained function. More powerful, less shareable.

            check_func :: Job -> Bool
                Ask for user input or do some other check to approve of a Job

            kptden :: Bool
                Replace kpt with kptden (average of kptden_x and kptden_y)

            emttol :: Float
                Used for a filter [Dict]->[Dict] that tries to remove redundant
                jobs by removing those with near-identical EMT energy.
                Use a negative number to turn off the filter
        """
        self.query      = query
        self.verbose    = verbose
        self.transform  = transform
        self.check_func = check_func

    def submit(self
              ,walltime           : typ.Optional[float]        = None
              ,clust_str          : typ.Optional[str]          = None
              ,nodes              : typ.Optional[int]          = None
              ) -> typ.Any:
        query_results  = self.query.query_dict()
        n,sub          = len(query_results),None
        if n == 0: print('No jobs meet your query: ',map(str,self.query.constraints))
        else:
            question = 'Do you want to modify %d jobs? \n(y/n)--> '%len(query_results)
            if not self.verbose or self._ask(question):
                made_jobs    = self._make_jobs(query_results)
                checked_jobs = list(filter(self.check_func,made_jobs))
                n            = len(checked_jobs)
                question     = 'Do you want to launch %d new jobs?\n(y/n)---> '%n
                if (not self.verbose) or (n < 50 and self._ask(question)) or (n >=50 and self._hard_ask(n)):
                    sub = {}
                    for j in checked_jobs:
                        print('not using graphs for equality checking')
                        sub.update(j.submit(repeats,walltime=walltime,clust_str=clust_str,nodes = nodes)) # USE GRAPH FOR EQUALITY TESTING
                    print('Launch with: catalog.misc.utilities.launch()')
                    return sub # return output of add_wf for last launch
                else: print('No jobs submitted')
            else: print('No jobs submitted')

    def _make_jobs(self
                  ,query_results        : typ.List[dict]
                  ) -> typ.List['Job']:
        ps = query_results #initialize with length 1 list of dictionaries
        for t in self.transform: ps = self._flatten(list(map(t,ps)))
        return [j for j in map(current_version.process_dict,ps) if not repeats.check_repeat(j)]

    def _ask(self, s : str) -> bool:      return 'y' in input(s).lower()
    def _hard_ask(self, n : int) -> bool: return input('Type {0} to launch {0} new jobs: '.format(n)) == str(n)
    def _flatten(self, lst : typ.List[typ.List[dict]])->typ.List[dict]:
        """
        Flatten a list of list of dictionaries. If there are non-list elements, treat them like singleton lists
        """
        output  = []
        for l in lst:
            if isinstance(l,list) and len(l)>0 and isinstance(l[0],dict): output.extend(l)
            elif len(l)==0: continue
            elif isinstance(l,dict): output.append(l)
            else: raise ValueError('transform functions should return Dict or [Dict]')
        return output

#################################################################################
# TRANSFORM FUNCS
#################################################################################

###################
# General utilities
#------------------
def identity(x:typ.Any) -> typ.Any : return x

def print_params(message : str ='printing params') -> transform_func_type:
    def f(d : dict) -> typ.List[dict]:
        print('\n\n%s\n\n%s\n\n'%(message,abbreviate_dict(d)))
        return [d]
    return f

def update_keys(modifying_dict : dict) -> transform_func_type:
    """update dictionary with dictionary of unary functions"""
    def f(d : dict) -> typ.List[dict]:
        return [{k:modifying_dict.get(k,identity)(v) for k,v in d.items()}]
    return f

def insert_keys(**kwargs : typ.Any) -> transform_func_type:
    """Insert/overwrite keys based on keyword arguments"""
    def f(d : dict) -> typ.List[dict]:
        return [merge_dicts([d,kwargs])]
    return f

def vary_dict(k     : str
             ,vs    : typ.List[typ.Any]
             ) -> transform_func_type:
    """create transform function to vary a parameter k over range vs"""
    if not isinstance(vs,list): vs=[vs]
    def f(d : dict) -> typ.List[dict]:
        ds  = []
        for v in vs:
            d[k]=v
            ds.append(d.copy())
        return ds
    return f


#####################
# Basic substitutions
#-------------------
def assign_parent(d : dict) -> typ.List[dict]:
    d['parent']=d.pop('storage_directory')
    return [d]

def job_name_kwargs(d : dict) -> typ.List[dict]:
    d['kwargs']=json.dumps({'job_name':d['job_name']})
    return [d]

def preserve_dwrat(d : dict) -> typ.List[dict]:
    if d.get('dw') is not None: d['dwrat']=d.pop('dw')/d['pw']
    return [d]

def preserve_kptden(d : dict) -> typ.List[dict]:
    d['kptden'] = np.mean([d.pop('kptden_x'),d.pop('kptden_y')])
    del d['kpts']
    return [d]


##################
# Create new trajs
#-----------------


def make_bare_slab(facet            : typ.Tuple[int,int,int]
                  ,xy               : typ.Tuple[int, int]
                  ,layers           : int
                  ,constrained      : typ.List[int]
                  ,symmetric        : bool
                  ,vacuum           : float
                  ) -> transform_func_type:
    def make(d : dict) -> typ.List[dict]:
        # AD HOC WARNING
        fudge_factor = 0.5 if d['structure'] in ['hcp','hexagonal'] else 1
        # AD HOC WARNING

        bulk = json_to_traj(d.pop('inittraj'))
        surface_dict = surfFuncs.bulk2surf(bulk,facet,xy,int(layers*fudge_factor),constrained,symmetric,vacuum)
        base_job_name = d['job_name']
        output_params = []
        for side,surf_ase_objs in surface_dict.items():
            for i, surf in enumerate(surf_ase_objs):
                d_curr = deepcopy(d)
                d_curr['inittraj'] = traj_to_json(surf)
                d_curr['job_name'] = base_job_name+'_{}'.format(surf.side)
                if len(surf_ase_objs)>1:
                    d_curr['job_name'] = d_curr['job_name']+'_{}'.format(i)
                output_params.append(d_curr)
        return output_params
    return make

def make_bare_all_facets(max_index        : int
                   ,xy               : typ.Tuple[int, int]
                   ,layers           : int
                   ,constrained      : typ.List[int]
                   ,symmetric        : bool
                   ,vacuum           : float
                   ) -> transform_func_type:
    def make(d : dict) -> typ.List[dict]:
        # AD HOC WARNING
        fudge_factor = 0.5 if d['structure'] in ['hcp','hexagonal'] else 1
        # AD HOC WARNING

        bulk = json_to_traj(d.pop('inittraj'))
        facets = surfFuncs.get_unique_facets(bulk,max_index)
        output_params = []
        for facet in facets:
            surface_dict = surfFuncs.bulk2surf(bulk,facet,xy,int(layers*fudge_factor),constrained,symmetric,vacuum)
            base_job_name = bare_name(facet,xy,layers)(d['job_name'])
            for side,surf_ase_objs in surface_dict.items():
                for i, surf in enumerate(surf_ase_objs):
                    d_curr              = deepcopy(d)
                    d_curr['facet']     = json.dumps(facet)
                    d_curr['inittraj']  = traj_to_json(surf)
                    d_curr['job_name']  = base_job_name+'_{}'.format(surf.side)
                    if len(surf_ase_objs)>1:
                        d_curr['job_name'] = d_curr['job_name']+'_{}'.format(i)
                    output_params.append(d_curr)
        return output_params
    return make


################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

####################
# Modify input trajs
#-------------------
def repeat_atoms(repeat            : typ.Tuple[int,int,int]
                ,fix_cell_problems : bool = False
                ,adjust_kpts       : bool = True
                ) -> transform_func_type:

    """ Helpful doc string"""
    def f(p : dict) -> typ.List[dict]:

        surface       = json_to_traj(p['inittraj'])
        if fix_cell_problems:
            surface       = fix_negative_unit_cells(surface)
            facet         = json.loads(p.get('facet'))
            surface       = reorient_z_ase(surface,facet)
        surface       *= repeat
        p['inittraj'] = traj_to_json(surface)

        #Adjust kpts if user specifies it
        if adjust_kpts:
            kpts          = json.loads(p['kpts'])
            kpts          = [int(k/rep) for rep,k in zip(repeat,kpts)]
            p['kpts']     = json.dumps(kpts)
        return [p]
    return f

def add_adsorbates(ads_list             : list
                  ,list_of_filter_funcs : typ.List[typ.Any] = []
                  ,constraint_dict      : dict              = {}
                  ,emttol               : float             = 0.001
                  ,site_type            : str               = 'all'
                  ,symm_reduce          : float             = 0.0
                  ,height               : float             = 2.0
                  ) -> transform_func_type:
    """ Helpful doc string"""
    from adsorber import filter_functions as ff
    from adsorber.objects.site import Site
    from adsorber.objects.asv import AdsorbateSiteVector

    if not isinstance(ads_list, list): ads = [ads_list]
    default_filters       = [ff.manual_filter(),ff.emt_filter(emttol=emttol)]
    list_of_filter_funcs += default_filters
    def f(p : dict) -> typ.List[dict]:
        import itertools
        new_params          = []
        surface             = json_to_traj(p['inittraj'])
        # surface             = fix_negative_unit_cells(surface)
        facet               = json.loads(p['facet'])
        # surface             = reorient_z_ase(surface,facet)


        #Remove the existing adsorbates to accurately find remaining adsorbate sites
        adsorbates          = json.loads(p['adsorbates']) if not p.get('adsorbates') is None else []
        ads_inds            = flatten(adsorbates)
        current_ads         = [surface[ind] for ind in ads_inds]
        surface_without_ads = remove_atoms(surface.copy(),ads_inds)

        #Get sites from pymatgen AdsorbateSiteFinder
        site_locs           = get_sites(surface_without_ads,facet, site_type = site_type, symm_reduce = symm_reduce, height = height)
        #remove any site within 2 covalent radii or 2 Angstroms of existing adsorbates
        site_locs           = filter(lambda pos: not any([collision(pos,atom,surface.cell,[1,1,1],dis_ind = 'xy') for atom in current_ads]),site_locs)
        sites               = map(lambda s: Site(surface_without_ads.copy(),s,facet),site_locs)
        #Create the AdsorbateSiteVector objects
        ads_site_pairs = itertools.product(ads_list,sites)
        asv_list = flatten([[AdsorbateSiteVector(a,s,v,constraint_dict = constraint_dict,additional_ads = current_ads) for v in s.vectors()] for a,s in ads_site_pairs])

        #Apply the filter functions to the sites
        print('Job Name: {0}'.format(p['job_name']))
        print('Applying Filter Functions to Adsorbate Structures:')
        for filter_func in list_of_filter_funcs:
            asv_list = filter_func(asv_list) # APPLY FILTERS

        #Get the modified param dicts from each of the remaining ASV objects
        for asv in asv_list:
            new_params.append(asv.get_modified_param_dict(p))
        return new_params
    return f

def constrain_slab(slab_constraint_dict : dict
                  ) -> transform_func_type:
    """
    apply a constraint to the inittraj in the params dictionary
    slab_constraint_dict :: dict
    the dictionary needs 3 fields
        'type'      :: ASE Constraint Function (i.e. ase.constraints.FixAtoms)
        'mask'      :: a function that takes in an Atom object and returns a boolean
        'inputs'    :: a dictionary with the inputs to the specific constraint (ie. {'direction': [0,0,1]} for FixedLine)
    """

    def get_atom_constraint(constraint_dict : dict
                           ,atom            : ase.Atom
                           ) -> typ.List[typ.Any]:
        if constraint_dict == {}:
            return []
        if constraint_dict['mask'](atom):
            if constraint_dict['type'].__name__ in ['FixedLine','FixedPlane']:
                return [constraint_dict['type'](a = atom.index, **constraint_dict['input'])]
            elif constraint_dict['type'].__name__ in ['FixAtoms']:
                return [constraint_dict['type'](indices = [atom.index])]
        return []

    def merge_FixAtoms_constraint(atoms : ase.Atoms) -> ase.Atoms:
        import ase.constraints as acon  # type: ignore
        other_constraints   = []        # type: list
        fixed_atom_indices  = []        # type: list
        for constraint in atoms.constraints:
            if 'FixAtoms' in constraint.__repr__():
                fixed_atom_indices = np.append(fixed_atom_indices,constraint.get_indices())
            else:
                other_constraints.append(constraint)
        unique_fixed_atom_indices = list(set(fixed_atom_indices))
        atoms.set_constraint([acon.FixAtoms(indices = unique_fixed_atom_indices)]+other_constraints)
        return atoms

    def f(p : dict) -> typ.List[dict]:
        #Check to see if any constraints need to be applied
        if slab_constraint_dict == None:
            return [p]
        bare      = json_to_traj(p['inittraj'])
        num_atoms = len(bare)
        con_bare  = bare.copy()
        #Constrain each atom if it matches the mask in slab_constraint_dict
        for atom in con_bare:
            con_bare.set_constraint(con_bare.constraints+get_atom_constraint(slab_constraint_dict, atom))
        #Merge the FixAtoms constraints together
        con_bare          = merge_FixAtoms_constraint(con_bare)
        p_new             = p.copy()
        p_new['inittraj'] = traj_to_json(con_bare)
        return [p_new]

    return f

def select_bonded_inds() -> transform_func_type:
    from ase.visualize import view
    import time
    def f(p : dict) -> typ.List[dict]:
        atoms_obj            = json_to_traj(p['inittraj'])
        view(atoms_obj)
        time.sleep(3)
        p_new                = p.copy()
        bonded_inds          = ask_for_inds(atoms_obj,'fbl',p)
        p_new['bonded_inds'] = json.dumps(bonded_inds)
        p_new['job_name']    = p['job_name'] + '_'.join(['','fbl',str(bonded_inds)])
        return [p_new]
    return f

def select_vacancy_index() -> transform_func_type:
    from ase.visualize import view
    import time
    def f(p : dict) -> typ.List[dict]:
        atoms_obj                 = json_to_traj(p['inittraj'])
        view(atoms_obj)
        time.sleep(3)
        output = []
        answer = 'y'
        while answer in 'y':
            p_new                = p.copy()
            vac_index            = ask_for_inds(atoms_obj,'vacancy',p_new)[0]
            p_new['vacancy_pos'] = json.dumps(list(atoms_obj[vac_index].position))
            p_new['inittraj']    = traj_to_json(remove_atom(atoms_obj,vac_index))

            #Modify the name of the job OLDNAME_SYMBOLvac_[X-Y-Z]
            #SYMBOL is chemical symbol of vacancy
            #X,Y,Z are the coordinates of vacancy atom
            vac_pos_str          = '['+'-'.join(map(str,list(atoms_obj[vac_index].position)))+']'
            symbol_str           = atoms_obj[vac_index].symbol+'vac'
            job_name_suffix      = '_'.join([symbol_str,vac_pos_str])
            p_new['job_name']    = '_'.join([p_new['job_name'],job_name_suffix])
            output.append(p_new)
            answer = input('Create another vacancy job (y/n)? ')
            while answer.lower() not in ['y','n']:
                answer = input('Create another vacancy job (y/n)? ')
        return output
    return f

def ask_for_inds(atoms_obj  : ase.Atoms
                ,job_type   : str
                ,p          : dict) -> typ.List[int]:
    assert job_type in ['fbl','vacancy']

    job_type_dict    = {'fbl'    :'Please use the atoms object shown to select the bonded indices'
                       ,'vacancy':'Please use the atoms object shown to select atom to delete'}

    number_to_select = {'fbl'    : 2
                       ,'vacancy': 1}

    question_dict    = {'fbl'    :'Please type the index of the {} bonded index: '
                       ,'vacancy':'Please type the index of the atom you want to delete: '}

    index_str = ['first','second','third','fourth']
    print('##############################################################')
    print('Job Name: {0}'.format(p['job_name']))
    print(job_type_dict[job_type])
    selected_inds = []
    for index in range(number_to_select[job_type]):
        answer        = -1
        while answer not in range(len(atoms_obj)):
            answer_str = input(question_dict[job_type].format(index_str[index]))
            try:
                answer = int(answer_str)
            except ValueError:
                answer = -1
                print('Please input an index in the atoms object')
        print('{0} selected as {1} index'.format(answer, index_str[index]))
        selected_inds.append(answer)
    print('{0} are the selected indices'.format(selected_inds))
    print('##############################################################')
    return selected_inds


################################################################
################################################################
################################################################



#######
# Other
#------
def get_vibids_if_needed(d : dict) -> typ.List[dict]:
    """Needs to be used before include_keys and rename_keys"""
    if d['job_type']!='vib':
        return [d]
    else:
        res = read_on_sher(d['storage_directory']+'/result.json')
        d['vibids']= json.loads(res)['vibids']
        return [d]

def vibrate_nonmetals(d : dict) -> typ.List[dict]:
    a = json_to_traj(d['inittraj'])
    nonmetal_inds = [i for i,x in enumerate(a.get_chemical_symbols()) if x in nonmetal_symbs]
    d['vibids']=json.dumps(nonmetal_inds)
    return [d]






##########################################
# Dictionaries for 'update_keys'
#---------------------------------------
def bare_name(facet     : typ.Tuple[int,int,int]
             ,xy        : typ.Tuple[int,int]
             ,layers    : int
             ) -> typ.Callable[[typ.Any],typ.Any]:
    s_f,s_xyz = map(str,facet),map(str,list(xy)+[layers])
    return lambda n: '_'.join([n,'-'.join(s_f),'x'.join(s_xyz)])

def replicate(x : int
             ,y : int
             ,z : int) -> typ.Callable[[typ.Any],typ.Any]:
    return lambda itp: pickle.dumps(pickle.loads(itp).repeat([x,y,z]))

#############################
# Interstital Method Archive
#############################

# def make_traj(d : dict) -> typ.List[dict]:
#     n,p_t,c,m,cs = list(map(d.pop,['numbers','posT','cell','magmoms','constrained']))
#     d['inittraj']= traj_to_json(make_atoms(n,p_t,c,m,cs))
#     return [d]


# def interstitial(constraint         : typ.List[typ.Any]     = []
#                 ,interstitial       : str                   = 'H'
#                 ,num                : int                   = 2
#                 ,limit              : typ.Optional[int]     = None
#                 ,emttol             : float                 =0.2
#                 ) -> GenData:
#     """
#     Specify an interstitial and the number (up to which) you want to add them to the cell
#     Filter duplicates via EMT-calculated energy (stipulate a tolerance)
#     Assume final structure is triclinic
#     """
#
#     default_constraint  = AND(LATTICEOPT_,QE_)
#     cnst               = AND(default_constraint,*constraint)
#     query=db.Query(make_traj+calc_args(drop=['xtol','kpts'])+[NICKNAME_CATALOG,JOBNAME
#                     ,As(STORDIR,'parent'),As(JOBTYPE,'jobkind')
#                     ,As(KPTDEN_X/2+KPTDEN_Y/2,'kptden')]
#                     ,cnst,group=JOB,table=tbl_relaxfinal_all,limit=limit)
#
#     return GenData(query,check_func = check_atoms
#                     ,transform=[insert_keys(jobkind='vcrelax',structure='triclinic')
#                                 ,make_traj
#                                 ,print_params()
#                                 ,interstitial_branch(interstitial,num)])
#
# def emt_filter(list_of_dictionaries : typ.List[dict]
#               ,tol                  : float             = 0.01
#               ) -> None:
#     """
#     Return *this* instead of a list_of_dictionaries at the end of
#     a transform function to filter out those with similar 'inittraj_pckl'
#     """
#     def extract(dic : dict) -> typ.Tuple[typ.Any,dict]:
#         atoms = pickle.loads(dic['inittraj_pckl'])
#         atoms.set_calculator(emt.EMT())
#         return (atoms.get_potential_energy(),dic)
#
# def interstitial_branch(inter           : str
#                        ,num             : int
#                        ,emttol          : float     = 0.2
#                        ) -> transform_func_type:
#     """
#     Given an interstitial (and how many), generate all structures
#     up to the number of desired interstitials
#     """
#     def f(p : dict) -> dict:
#         inittraj = pickle.loads(p['inittraj_pckl'])
#         name     = p['job_name']
#         spnpl    = any([x>0 for x in inittraj.get_initial_magnetic_moments()])
#         trajs = [[(inittraj,'')]] # initial state
#
#         for i in range(num):
#             lastround = trajs[-1]                # for all structures with n - 1 interstitials (note: initially there is only one, with 0 interstitials)
#             trajs.append([])                     # initialize container for all new structures with n interstitials
#             for inputtraj,strname in lastround:
#                 for newtraj,newname in interstitialFuncs.getInterstitials(inputtraj,inter,spnpl):
#                     trajs[-1].append((newtraj,strname+newname)) # add all new trajs found for all previous structures (append suffix to name)
#
#         def modParams(par : dict
#                      ,trj : ase.Atoms
#                      ,nam : str
#                      ) -> dict:                        # For a given new traj and name, create input parameters
#             p = copy.deepcopy(par)                     # All parameters common to all of these jobs
#             p['job_name']       += nam                 # Unique job name
#             p['inittraj_pckl']  = pickle.dumps(trj)    # Initial structure
#             return p
#
#         onelevel = [modParams(p,item[0],item[1]) for sublist in trajs[1:] for item in sublist] #collapse list of lists to a single list of input parameters
#
#         return emt_filter(onelevel)
#     return f






# def multiplyJSONlist(multiplier : float
#                     ) -> typ.Callable[[typ.Any],typ.Any]:
#     return {'kpts': lambda x: json.dumps(map(lambda y: multiplier*y,json.loads(x)))}
#

# Archived
# def multiplyMagmoms(multiplier : float):
#     def modTrajPckl(trajpckl : typ.Any) -> typ.Any:
#         atoms = pickle.loads(trajpckl)
#         atoms.set_initial_magnetic_moments(multiplier* atoms.get_initial_magnetic_moments())
#         return pickle.dumps(atoms)
#     return {'inittraj_pckl':modTrajPckl}

# def multiplyVacuum(val):
#     def modTrajPckl(trajpckl):
#         atoms = pickle.loads(trajpckl)
#         cellparams = cell_to_param(atoms.get_cell())
#         cellparams[2]*=val
#         atoms.set_cell(cellparams)
#         return pickle.dumps(atoms)
#     return {'inittraj_pckl':modTrajPckl}
#testing sync

# gpaw2QE = {'dftcode':lambda x: 'quantumespresso'
#             ,'psp':  lambda x: 'gbrv15pbe'
#             ,'xc':   lambda x: 'BEEF'
#             ,'dwrat':lambda x: 10}
#
# qe2VASP  = {'dftcode':lambda x: 'vasp'
#             ,'psp':  lambda x: '{}'
#             ,'xc':   lambda x: 'PBE'
#             } #ALL OTHER VASP PARAMTERS AT DEFAULT VALUES?


################################################################################
################################################################################

#
# def testNeb():
#     """
#     Does not fit into GenData paradigm, maybe use a superclass
#     """
#     from ase.io import read
#     params = {'inittraj_pckl': pickle.dumps(read('/scratch/users/ksb/test/initial.traj'))
#             ,'finaltraj_pckl': pickle.dumps(read('/scratch/users/ksb/test/final.traj'))
#             ,'jobkind':'neb','k': 0.3,'images':5,'structure':'triclinic'
#             ,'kwargs':json.dumps({'job_name':'vibTest'})
#             ,'parent': '/scratch/.../'
#             ,'parent2':'/scratch/...'
#             ,'dftcode':'quantumespresso' ,'xc':'PBE' ,'pw':500,'psp':'gbrv15pbe'
#             ,'kpts':json.dumps([6,6,6])
#             ,'fmax':0.05,'mixing':0.05,'nmix':10,'sigma':0.15,'nbands':-12,'dw':4000,'maxstep':400,'econv':0.0005,'mixingtype':'plain'}
#     jobs.Job(params).submit(repeats,walltime=1,nodes=5)
#     launch()

################################################################################
################################################################################
